#include "parse.h"
#include <grp.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/wait.h>

#define LOGFILE		"/var/log/pppcall" /* Where errors will be logged */
#define LOADABLE		/* #undef if you dont't have 'modprobe(1)' */
/* This is not a good place for this */
#define SERVER_ADDR	":192.88.125.2"

extern int set_defs(char *);
extern int check_locks(void);
extern int set_devices(char *);
extern int reset_locks(int);
extern int zap_locks();
extern int parse_sysfiles(char *, char *, int);
extern int set_speed(int, int);
extern char *say_speed(struct termios *);
extern int mem_read( char *, struct p_pntrs * );

extern struct defs *defs_base;

extern char pri_device[];
extern char alt_device[];
extern char pri_lock[];
extern char alt_lock[];
int debug;

void catch_int(int);
void catch_hup(int);
void catch_alarm(int);
void usage(void);
int load_ppp( void );
char *get_ip( char * );

/*
 * This is switch coding during development. Use #undef DO_PPP for just
 * logging in on remote machine
 */

#define DO_PPP

int
main(int argc, char *argv[])
{
	time_t now;
	char *p, *s;
	static char dest[] = SERVER_ADDR;
	char *system;
	char *host, *host_ip;
	char *bugger;
	int do_login;
	int time_out;
	int i, fd;

#ifdef DO_PPP
	int chld_pid;
#else
	int ch1, ch2, fd2;
	struct termios in_term, out_term;
#endif

	time_out = 60;
	host = dest;
	do_login = TRUE;
	debug = 0;
	system = NULL;
	i = 0;
	for( p = pri_device, s = pri_device + NAME_MAX;
		 p < s; *p++ = '\0' ); 
	for( p = alt_device, s = alt_device + NAME_MAX;
		 p < s; *p++ = '\0' ); 
	for( p = pri_lock, s = pri_lock + NAME_MAX;
		 p < s; *p++ = '\0' ); 
	for( p = alt_lock, s = alt_lock + NAME_MAX;
		 p < s; *p++ = '\0' ); 
	while (i < argc) {
		if (!(p = argv[i++]))
			break;
		if (*p == '-') {
			switch (*++p) {
			case 't':				  /* set the timeout interval */
				if (!*++p)
					p = argv[i++];
				time_out = atoi(p);
				break;
			case 'd':				  /* Fall through */
			case 'x':				  /* set the debug level */
				if (!*++p)
					p = argv[i++];
				debug = atoi(p);
				break;
			case 'h':					/* Use command line host */
				if (!*++p)
					p = argv[i++];
				host = p;
				break;
			case 'm':					/* do manual login */
				do_login = FALSE;
				break;
			case 's':					/* system name */
				if (!*++p)
					p = argv[i++];
				system = p;
				break;
			}
		} else if (*p == '?')
			usage();
	}
	if(( host_ip = get_ip( system )) == NULL ) {
		fprintf( stderr, "Unable to locate an IP address for '%s'.\n", system );
		host_ip = "";
	}
	if (getppid() <= 1) {
		freopen(LOGFILE, "a", stderr);
		setvbuf(stderr, NULL, _IONBF, 0);
		time(&now);
		fprintf(stderr, "%.24s: %s, Startup\n", ctime(&now), pri_device);
	}
	if (check_locks() != 0) {
		fprintf(stderr, "%.24s: %s locks are set.\n",
				  ctime(&now), pri_device);
		exit(-1);
	}
	if( system != NULL && load_ppp() ) {
		signal(SIGINT, catch_int);
		signal(SIGHUP, catch_hup);
		fd = parse_sysfiles(system, "ppp", do_login);
		if (fd > 0) {
for( i = 0; i < 6; i++ ) {
	char ch;

	read( fd, &ch, 1 );
	fprintf( stderr, "0x%.2x ", ch);
}
fputc( '\n', stderr );
			if(( chld_pid = fork()) == 0 ) {
				fclose(stderr);
				for (i = 0; i < fd; i++)
					close(i);
				for (i = 0; i < 3; i++)
					dup2(fd, i);
				if (fd >= 3) {
					close(fd);
					fd = 0;
				}
				if( debug ) bugger = "debug";
				else bugger = NULL;
				/* I don't want a device name in this list - why not?? */
				/* The FAQ & HOWTO says to use 'defaultroute. */
				/* 'crtscts', 'modem', 'disconnect' and net addr are in
				 * "/etc/ppp/options" */
 				execl("/usr/etc/pppd", "pppd", "defaultroute", host_ip,
						bugger, NULL);
				fprintf(stderr, "exec failed.\n");
			}
			reset_locks(chld_pid);
			fprintf(stderr,
					  "SUCCESS!! fd = %d, pppcall pid %d, pppd pid %d\n",
					  fd, getpid(), chld_pid);
			exit(0);
		}
	} else
		usage();
	exit(0);
}

void
usage()
{
	fprintf(stdout,
			  "uucall: -s system [-h host] [-t##] [-{d|x}]\n");
	fprintf(stdout,
			  "uucall: [-h] [-d] [-o | -e] systemname\n");
	exit(0);
}

void
catch_hup(int sig)
{
	fprintf(stderr, "Caught HANGUP, %d", sig);
	zap_locks();
	exit(-1);
}

void
catch_int(int sig)
{
	fprintf(stderr, "Caught SIGINT, %d", sig);
	zap_locks();
	exit(-1);
}

int
load_ppp()
{
#ifdef LOADABLE
	pid_t c_pid;

	if(( c_pid = fork()) == 0 ) {
		execl( "/sbin/modprobe", "modprobe", "ppp", NULL );
	}
	if( wait(NULL) == c_pid ) return( 1 );
	return( 0 ); /* a return of 0 is a failure */
#else
	return( 1 );
#endif
}

/*
 * This could be more elegant if I used 'gethostbyname(3N)' but it
 * doesn't return the address in a form recognized by 'pppd'.
 */

char *
get_ip( char *name )
{
	static char addr[16];
	struct p_pntrs host_p;
	char *p, *q, **pp;

	if( mem_read( HOSTFILE, &host_p ) < 0 ) return( NULL );
	for( pp = host_p.pp_pntr; *pp; pp++ ) {
		for( p = *pp, q = name; *p && *q; p++ ) {
			if( *p == '#' ) break;
			if( *p == *q ) q++;
			else q = name;
		}
		if( !*q ) {		/* This is the line we want */
			p = addr;
			*p++ = ':';
			q = *pp;
			while( *q != ' ' && *q != '\t' ) *p++ = *q++;
			*p = '\0';
			ppFREE( host_p );
			return( addr );
		}
	}
	ppFREE( host_p );
	return( NULL );
}
