#include <stdio.h>
#include <termios.h>
#include <fcntl.h>
#include <errno.h>
#include <sys/stat.h>
#include <syslog.h>
#include <sys/types.h>
#include <sys/ioctl.h>
#include <sys/dirent.h>
#include <signal.h>
#include <time.h>
#include <unistd.h>
#include <pwd.h>
#include <grp.h>
#include <string.h>
#include <stdlib.h>

#ifndef TRUE
#define TRUE 1
#endif
#ifndef FALSE
#define FALSE 0
#endif

#define UU_GRP						  /* Select appropiate parts of 'utils.c' */
#define UUDIR			"/usr/lib/uucp"
#define LCKDIR			"/var/lock/uucp"
#define DEFS_FILE		"/etc/gettydefs"
#define HOSTFILE		"/etc/hosts"
#define DEVDIR			"/dev"
#define CALLIN			"ttyS"	  /* The prefix for call in devices */
#define CALLOUT		"cua"		  /* The prefix for call out devices */

#define ROOT			"root"	  /* name of owner of tty device */
#define TTY				"tty"		  /* name of group of tty device */
#define UUCP_PID		30			  /* uid of "uucp" login from passwd */
#define UUCP_GID		14			  /* gid of "uucp" login from group */
#define ROOT_UID		0			  /* uid of "root" login from passwd */

/*
 * The 'gettydefs' file is parsed to this structure by
 * 'parse_gettydefs()'
 */
struct defs {
	char name[16];
	struct termios local;
	struct termios run;
	char prompt[80];
	char next[16];
};

enum how {
	nada,
	local,
	run,
	r_raw,
	cooked,
	toloc,
	notloc,
	nopar,
	par,
	echo,
	noecho,
	onlcr,
	bump,
	restart
};

/*
 * These were gleaned from Gert Doering's 'mfax' and a Sun's
 * 'termios.h'.  They should be in "linux/termios.h".
 */
#define CNULL		0				  /* Zero is Null, very profound. */
#define CINTR		'\003'		  /* ETX, ^C */
#define CQUIT		'\034'		  /* FS,  ^\ or ^| */
#define CERASE		'\010'		  /* BS,  ^H or \b */
#define CKILL		'\025'		  /* NAK, ^U */
#define CEOF		'\004'		  /* EOT, ^D */
#define CSWITCH	'\032'	/* SUB, ^Z */ /* For shl layers */
#define CSTART		'\021'		  /* DC1, ^Q */
#define CSTOP		'\023'		  /* DC3, ^S */
#define CSUSP		'\032'		  /* SUB, ^Z */
#define CEOL		'\0'			  /* NUL, ^@ */
#define CRPRNT		'\022'		  /* DC2, ^R */
#define CFLUSH		'\017'		  /* SI,  ^O */
#define CWERASE	'\027'		  /* ETB, ^W */
#define CEOL2		'\0'			  /* NUL, ^@ */
/* These don't seem to be used */
#define CLNEXT		0
#define CESC		'\034'		  /* FS,  ^\ or ^| */
#define CDEL		'\0377'		  /* ESC, why is this 0177 in my book? */
#define CEOT		'\004'		  /* EOT, ^D */
#define CNSWITCH	0				  /* We don't run 'shl'  */
#define CDSUSP		0

/*
 * This is what 'stty' thinks SANE should be.
 */
#define SANE_iflag BRKINT | ICRNL	/* | IMAXBEL */
#define SANE_oflag OPOST  | ONLCR
#define SANE_cflag CREAD
#define SANE_lflag ISIG | ICANON | ECHO | ECHOK | ECHOCTL | ECHOKE | IEXTEN
