#include "parse.h"
#include <setjmp.h>

extern void display_char(unsigned char *, unsigned char);

jmp_buf exp;

int chat(int, char *, char, char *, char *);
char *digest(int, char *, char, char *, char *);
char *translate(char *, int, char *);
int send(char *, int);
void *exp_catch(void);
int expect(char *, int);
int check_uid(char *, int);
extern int set_speed(int, enum how);
extern int mpause(int m);
extern int debug;

/*
 * This routine accepts a 'chat' scripts from a Devices file or a
 * Systems file.
 *   fd:     file descriptor for device we want to chat with.
 *   script: the script to be acted upon.
 *   flag:   is a '\D', \T or NUL as read as part of the fifth
 *           field of the Devices file entry.  These are analogous
 *           to the '\D' or '\T' which are part of Dialers chat script.
 *   phnum:  extracted from the fifth field of the Systems file stating
 *           the phone number to be dialed.  It may contain references
 *           to entries in the Dialcodes file.  If it does, '\D' must
 *           be used in the Dialers script or appended as ',\T' to
 *           Dialer's reference, the fifth field of the Devices file.
 *   subs:   these are the second field of the Dialer's entry.  They
 *           substitute a '=' to the actual symbol needed by the modem
 *           to sense a secondary dial tone and a '-' to a modem symbol
 *           for pause.
 * 
 * These are then passed on to the 'digest' routine for action.
 */
int
chat(int fd, char *script, char flag, char *phnum, char *subs)
{
	char wrk[2 * INBUF];
	char *pnt[20];
	char *p, *q, *r;
	char *data;
	char ch1, ch2;
	int count, ret, tries, secure;

	if ((debug > 2) && (getuid() == ROOT_UID))
		fprintf(stderr, "chat:\t>%s<\n", script);
	tries = 0;
 restart:
	if (debug > 2)
		fprintf(stderr, "restart: tries %d.\n", tries);
	if (tries++ > 2)
		return (-1);
	secure = 0;
	ch1 = '\0';
	ret = -1;
	p = script;
	q = wrk;
	r = wrk + sizeof(wrk);

	/* 
	 * count is an alternator for switching between 'expect' and 'send'
	 */
	for (count = 0; count < 20; count++) {
		pnt[count] = q;
		/* Odd counts are 'expect's and even counts are 'send's */
		if (count & 0x01)
			ch2 = ch1;
		else
			ch2 = '-';
		while (*p && *p != ' ' && *p != '\t' && *p != '\n'
				 && *p != ch2 && q < r)
			*q++ = *p++;
		ch1 = *p++;
		*q++ = '\0';
		if (secure != count)
			secure = check_uid(pnt[count], count);
		if(( data = digest(fd, pnt[count], flag, phnum, subs)) == NULL ) {
			return( -1 );
		}
		if (count & 0x01) {
			if (debug > 2) {
				fprintf(stderr, "send: ");
				if (secure == count)
					display_char("********\n", '\0');
				else
					display_char(data, '\0');
				putc('\n', stderr);
			}
			ret = send(data, fd);
		} else {
			/* 
			 * expect() returns:
			 *  0 - sucessfully read expect string
			 *  1 - time out during 'expect'
			 *  2 - ERROR from modem
			 *  3 - called line BUSY
			 *  4 - framing error from modem
			 * -1 - system read() error
			 */
			ret = expect(data, fd );
			if( ret == 0 ) {
				if(  ch1 == '-') {
					/* Advance pointer past alternative 'expect' string */
					while (*p != ' ' && *p != '\t')
							p++;
				}
			}
			else if( ret == 1 && ch1 == '-' ) continue;
			else if( ret == 2 )
					goto restart;
			/* Send the 'ret' value back to parse() to retry or quit */
			else return( ret );
		}
		while (*p == ' ' || *p == '\t')
			p++;
		if (!ch1 || ch1 == '\n') {
			if (debug > 2)
				fprintf(stderr, "chat end: >%s< %d\n", pnt[count], ret);
			break;
		}
	}
	return (ret);
}

#define DUMP_BUF(r) \
   if( nl ) *q++ = '\n'; \
	*q = '\0'; \
	if( debug > 2 ) { \
		fprintf( stderr, "send: " ); \
		display_char( buf, '\0' ); \
		fprintf( stderr, "<%s>\n", r ); \
	} \
	send( buf, fd ); \
	q = buf;

/*
 * This routine scans 'strg' for all the special symbols listed in the
 * BNU uucp documentation.  Appropiate substitutions are made or actual
 * commands are performed.
 */
char *
digest(int fd, char *strg, char flag, char *phnum, char *subs)
{
	char *p, *q, *r, *s, *t;
	char *msg;
	static unsigned char buf[INBUF];
	int nl, i, num;

	if (debug > 3)
		fprintf(stderr, "digest: '%c', '%s', %s, %s\n", flag, subs, phnum, strg);
	p = strg;
	q = buf;
	r = buf + sizeof(buf);
	nl = 1;
	while (*p && q < r) {
		for (i = 0; i < 2; i++) {
			switch (i) {
			case 0:
				msg = "EOT";
				break;
			case 1:
				msg = "BREAK";
				break;
			default:
				msg = "";
			}
			s = p;
			t = msg;
			while (*s == *t)
				s++, t++;
			if (!*t)
				break;
		}
		if (i == 0) {
			*q++ = '\004';
			if (debug > 2)
				fprintf(stderr, "<EOT>");
		} else if (i == 1) {
			DUMP_BUF("break");
			ioctl(fd, TCSBRK, 0);
		} else {
			if (*p == '\\') {
				switch (*++p) {
				case 'b':			  /* \b -> backspace */
					*q++ = '\b';
					break;
				case 'c':			  /* \c -> suppress new line */
					nl = 0;
					break;
				case 'd':			  /* \d -> delay */
					DUMP_BUF("delay");
					sleep(1);
					break;
				case 'K':			  /* \K -> send break */
					DUMP_BUF("break");
					ioctl(fd, TCSBRK, 0);
					break;
					break;
				case 'n':			  /* \n -> send new line */
					nl = 1;
					break;
				case 'N':			  /* \N -> send NUL */
					*q++ = '\0';
					break;
				case 'p':			  /* \p -> pause */
					DUMP_BUF("pause");
#ifdef HAVENAP
					nap(25);			  /* 25 milli-seconds */
#else
					mpause(25);
#endif
					break;
				case 'r':			  /* \r -> send carriage return */
					*q++ = '\r';
					break;
				case 's':			  /* \s -> send space */
					*q++ = ' ';
					break;
				case 't':			  /* \t -> send tab */
					*q++ = '\t';
					break;
				case '\\':			  /* \\ -> send backslash */
					*q++ = '\\';
					break;
				case 'M':			  /* \M -> set termio LOCAL */
					set_speed(fd, toloc);
					if (debug > 2)
						fprintf(stderr, "<local>");
					break;
				case 'T':			  /* \T -> translate dialcodes */
					if (phnum == NULL)
						break;
					if (flag == '\0')
						flag = 'T';
					goto copy;
				case 'D':			  /* \D -> don't translate dial */
					if (phnum == NULL)
						break;
					if (flag == '\0')
						flag = 'D';
				 copy:
					if((s = translate(phnum, flag, subs)) == NULL ) {
						fprintf( stderr, "Error parsing 'Dialcodes'.\n" );
						return( NULL );
					}
					while (*s)
						*q++ = *s++;
					break;
				case 'm':			  /* \m -> set termio ~LOCAL */
					set_speed(fd, notloc);
					if (debug > 2)
						fprintf(stderr, "<run>");
					break;
				case '0':
					num = 0;			  /* \ddd -> convert octal ascii to int */
					while (*p >= '0' && *p <= '7') {
						num *= 8;
						num += *p - '0';
						p++;
					}
					p--;
					*q++ = num;
					break;
				}
			} else
				*q++ = *p;
		}
		p++;
	}
	if (nl)
		*q++ = '\n';
	*q = '\0';
	return (buf);
}

/*
 * This will translate Dialcodes entries when \T is used in a chat
 * script or if ,\T is appended to the Dialers ( fifth ) field of the
 * Devices entry.
 */
char *
translate(char *phnum, int flag, char *subs)
{
	FILE *codes;
	char path[NAME_MAX];
	static char out[SHRT_BUF];
	char data[INBUF];
	char aka[SHRT_BUF];
	char *p, *q, *r, *s, *t, *u;
	char *end;

	if (debug > 3)
		fprintf(stderr, "translate: '%c', '%s', %s\n", flag, subs, phnum);
	p = phnum;
	q = out;
	s = out + sizeof(out);
	if (flag == 'T') {
		strcat(strcat(strcpy(path, UUDIR), "/"), DIALCODES);
		if ((codes = fopen(path, "r")) != NULL) {
			r = aka;
			t = aka + sizeof(aka);
			while (*p >= 'a' && *p <= 'z' && r < t)
				*r++ = *p++;
			*r = '\0';
			while ((end = fgets(data, sizeof(data), codes)) != NULL) {
				r = aka;
				u = data;
				while (*r == *u)
					r++, u++;
				if (!*r) {
					while (*u == ' ' || *u == '\t')
						u++;
					while ((*u && *u >= '0' && *u <= '9')
							 || *u == '=' || *u == '-') {
						if (*u == '=' && subs[1] != ',')
							*q++ = subs[1];
						else if (*u == '-' && subs[3] != ',')
							*q++ = subs[3];
						else
							*q++ = *u;
						u++;
					}
					break;
				}
			}
			fclose(codes);
		} else if (debug > 2)
			fprintf(stderr, "Unable to open '%s', %s\n", path,
					  sys_errlist[errno]);
	}
	while (*p == ' ' && *p == '\t')
		p++;
	while ((*p && *p >= '0' && *p <= '9') || *p == '=' || *p == '-') {
		if (*p == '=' && subs[1] != ',')
			*q++ = subs[1];
		else if (*p == '-' && subs[3] != ',')
			*q++ = subs[3];
		else
			*q++ = *p++;
	}
	*q = '\0';
	return( *p ? NULL : out );
}

int
send(char *strg, int fd)
{
	unsigned char *p;

	p = strg;
	while (*p) {
		if (write(fd, p++, 1) <= 0)
			return (-1);
	}
	write(fd, '\0', 1);
	return (0);
}

void *
exp_catch(void)
{
	longjmp(exp, 1);
}

int
expect(char *strg, int fd)
{
	char *p, *q, *r;
	char ch1, ch2, ch3;
	static char err[] = "ERROR";
	static char busy[] = "BUSY";
	int ret;

	ret = 0;
	if (debug > 2) {
		p = strg;
		while (*p)
			p++;
		p--;
		fprintf(stderr, "expect: %s%s", strg, *p == '\n' ? "" : "\\c\n");
		fprintf(stderr, "got: ");
	}
	if (*strg == '"' && *(strg + 1) == '"') {
		if (debug > 2)
			fputs("\"\"\n", stderr);
	} else {
		if (setjmp(exp)) {
			fprintf(stderr, "TIMED OUT expecting '%s'.\n", strg );
			return (1);
		}
		signal(SIGALRM, (void *) exp_catch);
		alarm(30);
		ret = 0;						  /* Now let longjmp try to clobber it */
		p = strg;
		q = err;
		r = busy;
		ch2 = '\0';
		ch3 = '\0';
		while (*p && *q && *r && *p != '\n' && *p != '\r') {
			if (read(fd, &ch1, 1) < 0) {
				ret = -1;
				break;
			}
			if (ch1 == *p)
				p++;
			else {
				p = strg;
				if (ch1 == *p)
					p++;
			}
			if (ch1 == *q)
				q++;
			else
				q = err;
			if (ch1 == *r)
				r++;
			else
				r = busy;
			if (debug > 2) {
				/* Just to avoid printing out a multitude of ^M's and ^J's */
				if ((ch1 == '\n' && ch2 == '\n') ||
					 (ch1 == '\r' && ch2 == '\r'))
					continue;
				display_char(NULL, ch1);
			}
			if (ch3 == '\377' &&
				 (ch2 == '\0' || ch2 == '\377') &&
				 (ch1 == '\0' || ch1 == '\377')) {
				putc('\n', stderr);
				alarm(0);
				return (4); /* Framing Error */
			}
			ch3 = ch2;
			ch2 = ch1;
		}
		if (!*q) {
			fputs(" got ERROR.\n", stderr);
			ret = 2;					  /* ERROR from modem */
		}
		else if (!*r) {
			fputs(" got BUSY.\n", stderr);
			ret = 3;					  /* BUSY from modem */
		}
		else if (ret < 0)
				fputs(" blew it.\n", stderr);
		else if( debug > 2 )
				fputs(" got it.\n", stderr);
		alarm(0);
	}
	return (ret);
}

/*
 * This checks the user id.  If it is not 'root', the string is checked
 * for key symbols and returns the count integer if the current send
 * sequence should be masked as in when sending the phone number.  It
 * returns the next higher integer if the expect string is 'ogin:' or
 * 'sword:'.
 */
int
check_uid(char *strg, int count)
{
	char *p, *q, *s, *msg;
	int i;

	if (getuid() != ROOT_UID) {
		for (i = 0; i < 4; i++) {
			switch (i) {
			case 0:
				msg = "\\T";
				break;
			case 1:
				msg = "\\D";
				break;
			case 2:
				msg = "in:";
				break;
			case 3:
				msg = "rd:";
				break;
			default:
				msg = "";
				break;
			}
			p = strg;
		 loop:
			q = msg;
			while (*p && *p != *q)
				p++;
			if (*p) {
				s = p;
				while (*p && *q && *p == *q)
					p++, q++;
				if (*q) {
					p = ++s;
					goto loop;
				} else if (i < 2)
					return (count);
				else
					return (count + 1);
			}
		}
	}
	return (0);
}
