#include "parse.h"

extern int open_device(char *);
extern int check_time(char *);
extern int chat(int, char *, char, char *, char *);
extern int pntr_init(struct p_pntrs *, int);
extern int add_on(struct p_pntrs *, char * );
extern int read_defs(char *, char *);
extern int set_speed(int, enum how);
extern int zap_locks();

extern int debug;
extern struct p_pntrs NULLpntrs;

int parse_sysfiles(char *, char *, int);
int parse_file(struct p_pntrs *, struct p_pntrs *, char *);
int getline(FILE *, char *, char *, int);

int
parse_sysfiles(char *node, char *prog, int do_login)
{
	FILE *sys_fd;
	char path[NAME_MAX];
	char data[INBUF];
	char info[SHRT_BUF];
	char flag, grade, proto;
	char *sys[6];
	char *devs[5];
	char *dials[4];
	char *p, *q, *r, *s, *t, *wrd;
	char **pp, **qq, **rr;
	struct p_pntrs systems, devices, dialers;
	struct p_pntrs sys_info, dev_info, dial_info;
	int i, found, ret, pass, secure;
	int dial_fd, run_fd;
	int ioctl_flags;

	if (getuid() == ROOT_UID)
		secure = 1;
	else
		secure = 0;

	if (pntr_init(&systems, 0) < 0) {
		fprintf(stderr, "Unable to initalize systems pointers.\n");
		return (-1);
	}
	if (pntr_init(&devices, 0) < 0) {
		fprintf(stderr, "Unable to initalize devices pointers.\n");
		ppFREE( systems );
		return (-1);
	}
	if (pntr_init(&dialers, 0) < 0) {
		fprintf(stderr, "Unable to initalize dialers pointers.\n");
		ppFREE( systems );
		ppFREE( devices );
		return (-1);
	}

	/*
	 * Open the 'Sysfiles' file.  It and 'gettydefs' and 'Dialcodes' are
	 * the only files with predefined names.  The names for the
	 * 'systems', 'devices' and 'dialers' type files are defined here.
	 */
	p = path;
	q = UUDIR;
	s = p + sizeof(path);
	while (*q && p < s)
		*p++ = *q++;
	*p++ = '/';
	q = SYSFILES;
	while (*q && p < s)
		*p++ = *q++;
	*p = '\0';
	if (debug > 2)
		fprintf(stderr, "Opening '%s'\n", path);
	if ((sys_fd = fopen(path, "r")) == NULL) {
		fprintf(stderr, "Unable to open '%s'\n", path);
		ppFREE( systems );
		ppFREE( devices );
		ppFREE( dialers );
		return (-1);
	}
	found = FALSE;
	run_fd = -1;
	ret = 0;
	while (fgets(data, sizeof(data), sys_fd) > 0) {
		p = data;
		s = p + sizeof(data);
		while (*p == ' ' || *p == '\t' || *p == '\n')
			p++;
		if (*p == '#' || !*p)
			continue;
if( debug > 2 )
	fprintf( stderr, "Found '%s'\n", data );
		for (i = 0; i < 4; i++) {
			while (*p == ' ' || *p == '\t' || *p == '\n')
				p++;
			switch (i) {
			case 0:
				wrd = "service";
				break;
			case 1:
				wrd = "systems";
				break;
			case 2:
				wrd = "devices";
				break;
			case 3:
				wrd = "dialers";
				break;
			default:
				wrd = "";
				break;
			}
			r = wrd;
			while (*p == *r && p < s)
				p++, r++;
			if (!*r) {
				/* loop if we have multple colon separated entries */
			 loop:
				while ((*p == '=' || *p == ' ' || *p == '\t'
						  || *p == ':' || *p == '\n') && p < s)
					p++;
				if (i && found) {
					r = info;
					t = r + sizeof(info);
					while (*p && *p != ' ' && *p != '\t' && *p != '\n'
							 && *p != ':' && *p != '\\' && r < t)
						*r++ = *p++;
					*r = '\0';
				}
				switch (i) {
				case 0:
					r = prog;
					while (*p == *r)
						p++, r++;
					if (!*r)
						found = TRUE;
					break;
				case 1:
					add_on(&systems, info );
					break;
				case 2:
					add_on(&devices, info );
					break;
				case 3:
					add_on(&dialers, info );
					break;
				default:
					break;
				}
			}
			if (i == 0 && found == FALSE)
				break;
			while (*p == ' ' || *p == '\t')
				p++;
			if (*p == ':')
				goto loop;
			if (*p == '\\') {
				fgets(data, sizeof(data), sys_fd);
				p = data;
			}
		}
	}
	fclose(sys_fd);

	/* Set defaults */
	if (systems.ch_size == 0)
		add_on(&systems, "Systems" );
	if (devices.ch_size == 0)
		add_on(&systems, "Devices" );
	if (dialers.ch_size == 0)
		add_on(&systems, "Dialers" );

	/* Now gather the info. */
	flag = '\0';
	if (pntr_init(&sys_info, 0) < 0) {
		fprintf(stderr, "Unable to initalize 'system' pointers.\n");
		ppFREE( systems );
		ppFREE( devices );
		ppFREE( dialers );
		return (-1);
	}
	if (debug > 0)
		fprintf(stderr, "Systems entry '%s' wanted:\n", node);
	if ((ret = parse_file(&systems, &sys_info, node)) < 0) {
		fprintf(stderr, "Error in reading 'Systems' file.\n");
		ppFREE( systems );
		ppFREE( devices );
		ppFREE( dialers );
		ppFREE( sys_info );
		return (-1);
	}
	pass = 0;
	for (pp = sys_info.pp_pntr; *pp; pp++) {
		p = *pp;

		if (debug > 2 && secure)
			fprintf(stderr, "Systems >%s<\n", p);
		/* Load the sys[] pointer array */
		for (i = 0; i < 5; i++) {
			while (*p == ' ' || *p == '\t')
				p++;
			sys[i] = p;
			while (*p != ' ' && *p != '\t' && *p != '\n')
				p++;
			*p++ = '\0';
		}
		/* Put rest of line in the fifth buffer */
		sys[i] = p;
		while (*p)
			p++;

		/* Extract 'grade' if given */
		p = sys[1];
		while (*p && *p != '/')
			p++;
		if (*p == '/') {
			*p++ = '\0';
			grade = *p;
		} else
			grade = '\0';

		/* Extract 'protocol' if given. */
		p = sys[2];
		while (*p && *p != ',')
			p++;
		if (*p == ',') {
			*p++ = '\0';
			proto = *p;
		} else
			proto = '\0';

		/* Might as well check the time now. */
		if ((check_time(sys[1])) < 0) {
			fprintf(stderr, "Wrong time to call\n");
			continue;
		}

		/*
		 * Now load the various 'Devices' references
		 */
		if ((pntr_init(&dev_info, 0)) < 0) {
			fprintf(stderr, "Unable to initalize 'devices' pointers.\n");
			ret = -1;
			goto quit2;
		}
		if (debug > 1)
			fprintf(stderr, "Device entry '%s' wanted:\n", sys[2]);
		if ((parse_file(&devices, &dev_info, sys[2])) < 0) {
			fprintf(stderr, "Unable to locate an entry for '%s' in '%s'.\n",
					  sys[2], devices.ch_pntr);
			ret = -1;
			goto quit2;
		}
		dial_fd = -1;				  /* to keep gcc quiet */
		for (qq = dev_info.pp_pntr; *qq; qq++) {
			/* Let the modem and system settle between passes. */
			if (pass++ > 0)
				sleep(5);
			p = *qq;
			if (debug > 2 && secure)
				fprintf(stderr, "Devices:\n>%s<\n", p);
			if( *p == '\0' ) continue;
			for (i = 0; i < 4; i++) {
				while (*p == ' ' || *p == '\t')
					p++;
				devs[i] = p;
				while (*p && *p != ' ' && *p != '\t' && *p != '\n')
					p++;
				if (!*p || *p == '\n')
					devs[++i] = NULL;
				*p++ = '\0';
			}
			/* Put rest of line in fourth buffer */
			while (*p == ' ' || *p == '\t')
				p++;
			devs[i] = p;
			while (*p)
				p++;

			/* Check for alternative location to specify protocol. */

			/*
			 * This is a poor alternative for what should really be done.
			 */
			p = devs[0];
			while (*p && *p != ',')
				p++;
			if (*p == ',') {
				*p++ = '\0';

				/*
				 * We give the protocol option in the Systems file
				 * priority.
				 */
				if (proto == '\0')
					proto = *p;
			}

			/*
			 * Check for Translation flag.  This seems redundant since a
			 * \T or \D in the chat script does the same thing.
			 */
			p = devs[4];
			while (*p && *p != ' ' && *p != '\t' && *p != '\n')
				p++;
			*p++ = '\0';
			while (*p && (*p == ' ' || *p == '\t'))
				p++;
			if (*p++ == '\\')
				flag = *p;

			/*
			 * I can't think of any application which would use multiple
			 * Dialers entries of the same name, but I'll include that
			 * capability anyhow.
			 */
			if ((pntr_init(&dial_info, 0)) < 0) {
				fprintf(stderr, "Unable to initalize 'dialers' pointers.\n");
				ret = -1;
			}
			if (debug > 1)
				fprintf(stderr, "Dialer entry '%s' wanted:\n", devs[4]);
			if ((parse_file(&dialers, &dial_info, devs[4])) < 0) {
				fprintf(stderr, "Unable to locate an entry for '%s' in '%s'.\n",
						  devs[4], dialers.ch_pntr);
				ret = -1;
			}
			for (rr = dial_info.pp_pntr; *rr; rr++) {
				p = *rr;
				if (debug > 2 && secure)
					fprintf(stderr, "Dialers:\n>%s<\n", p);
				if( *p == '\0' ) continue;
				for (i = 0; i < 2; i++) {
					while (*p == ' ' || *p == '\t')
						p++;
					dials[i] = p;
					while (*p != ' ' && *p != '\t' && *p != '\n')
						p++;
					*p++ = '\0';
				}
				/* Put rest of line in third buffer */
				while (*p == ' ' || *p == '\t')
					p++;
				dials[i] = p;
				while (*p)
					p++;
				if (debug > 3) {
					for (i = 0; i < 6; i++)
						fprintf(stderr, "sys[%d]: >%s<\n", i, sys[i]);
					for (i = 0; i < 5; i++)
						fprintf(stderr, "devs[%d]: >%s<\n", i, devs[i]);
					for (i = 0; i < 3; i++)
						fprintf(stderr, "dials[%d]: >%s<\n", i, dials[i]);
				}

				/*
				 * devs[1] is the name of the 'call' device. devs[2] is the
				 * name of the 'dialer' device. devs[3] is the speed for
				 * dialer sequence.
				 */

				if (*devs[2] == '-')
					devs[2] = devs[1];
				if (debug > 1)
					fprintf(stderr, "Device port '%s' wanted:\n", devs[2]);
				if ((dial_fd = open_device(devs[2])) < 0) {
					fprintf(stderr, "Unable to open '%s' as dialer.\n", devs[2]);
					ret = -1;
					goto quit1;
				}

				if ((read_defs(devs[3], NULL)) < 0) {
					ret = -1;
					goto quit1;
				}
				set_speed(dial_fd, local);
				if (debug > 2)
						fprintf(stderr, "Toggling DTR-RTS.\n");
				/* Toggle DTR just for laughs */
				ioctl_flags = TIOCM_DTR | TIOCM_RTS;
				/* Turn off flags */
				if ((ret = ioctl(dial_fd, TIOCMBIC, &ioctl_flags)) < 0) {
					fprintf(stderr, "Unable to drop DTR and RTS, %s",
							  sys_errlist[errno]);
					goto quit1;
				}
				sleep(1);
				/* Turn on flags */
				if ((ret = ioctl(dial_fd, TIOCMBIS, &ioctl_flags)) < 0) {
					fprintf(stderr, "Unable to turn on DTR and RTS, %s",
							  sys_errlist[errno]);
					goto quit1;
				}
				set_speed(dial_fd, noecho);
				set_speed(dial_fd, nopar);
				/* 
				 * chat() returns:
				 *  0 - sucessfully established called out
				 *  1 - time out during 'expect'
				 *  2 - ERROR from modem (trapped by chat())
				 *  3 - called line BUSY
				 *  4 - framing error from modem
				 * -1 - system read() error
				 */
				while(( ret = chat(dial_fd, dials[2], flag, sys[4],
										 dials[1] )) == 4 ) {
					/* Try to autobaud the computer */
					if ((ret = set_speed(dial_fd, bump)) <= 0) {
						fprintf(stderr, "All available baud rates failed.\n");
						close( dial_fd );
						dial_fd = -1;
						break;
					}
				}
				if (ret == 3 || ret == 0)
					break;
			}
			/* We have working connection */
			if (ret == 0) {
				if ((ret = strcmp(devs[3], sys[3])) != 0) {
					/* We want to use a different gettydef for communication */
					if ((ret = read_defs(sys[3], NULL)) < 0)
						goto quit1;
				}
				if (strcmp(devs[1], devs[2]) == 0)
					run_fd = dial_fd;
				else {
					/* We want to use a different port for communication */
					if ((run_fd = open_device(devs[1])) < 0) {
						fprintf(stderr, "Error opening '/dev/%s'\n", devs[1]);
						ret = -1;
					}
					fprintf(stderr, "close dial_fd.\n");
					close(dial_fd);
				}
				dial_fd = -1;

				/* Run this only for automatic logins */
#if 0
				set_speed(run_fd, run );
/* set_speed(run_fd, noecho ); */
#endif
{
struct termios term = {
		0,
		0,
		B19200 | CS8 | CREAD | HUPCL,
		0,
		N_TTY,
		{ 0x03, 0x1c, 0x08, 0x15, 0x04, 0x00, 0x00, 0x00, 0x11,
				  0x13, 0x1a, 0x00, 0x12, 0x0f, 0x17, 0x00, 0x00 },
	};
ioctl( dial_fd, TCSETSW, &term );
}
				if (do_login) {
					if ((ret = chat(run_fd, sys[5], '\0', NULL, NULL)) != 0) {
						close(run_fd);
						run_fd = -1;
						ret = -1;
					}
				}
			}
		 quit1:
			ppFREE(dial_info);
			if (ret == 0)
				break;
			if( dial_fd >= 0 ) {
				close( dial_fd );
				dial_fd = -1;
			}
		}
	 quit2:
		ppFREE(dev_info);
		if (ret == 0)
			break;
	}
	ppFREE(sys_info);
	ppFREE(dialers);
	ppFREE(devices);
	ppFREE(systems);
	return (ret < 0 ? -1 : run_fd);
}

/*
 * Open each file listed in the p_pntrs 'file' as read from Sysfiles.
 * Scan for an entry with 'targ' as the first word on the line. Push
 * the returned line to the p_pntrs 'data' buffer. The line may be
 * extened with a backslash.
 */

int
parse_file(struct p_pntrs *names, struct p_pntrs *data, char *targ)
{
	FILE *sys_fd;
	char path[NAME_MAX];
	char line[INBUF];
	char *p, *q, *r;
	char **pp;
	int ret, found;

	ret = -1;
	found = -1;
	for (pp = names->pp_pntr; *pp; pp++) {
		/* Open the 'names' file */
		p = path;
		r = path + sizeof(path);
		q = UUDIR;
		while (*q && p < r)
			*p++ = *q++;
		*p++ = '/';
		q = *pp;
		while (*q && p < r)
			*p++ = *q++;
		*p = '\0';
		if (debug > 2)
			fprintf(stderr, "\tSearching '%s' for '%s'\n", path, targ );
		if ((sys_fd = fopen(path, "r")) == NULL) {
			fprintf(stderr, "Unable to open '%s'\n", path);
			return (-1);
		}
		/* Now read in all the lines starting with 'targ'. */
		while(( ret = getline(sys_fd, targ, line, sizeof(line) ))) {
			found = 0;
			add_on(data, line );
		}
		fclose(sys_fd);
	}
	return (found);
}

int
getline(FILE * sys_fd, char *targ, char *data, int size)
{
	char line[INBUF];
	char *p, *q, *r, *s;
	int found;

	if (*targ == '\0')
		return (-1);
	found = FALSE;
	if( debug > 4 )
			fprintf( stderr, "getline: targ '%s'\n", targ );
	while (!found && (fgets(line, sizeof(line), sys_fd) > 0)) {
		if( debug > 4 )
				fprintf( stderr, "getline: '%s'\n", line );
		p = line;
		r = line + sizeof(line);
		while ((*p == ' ' || *p == '\t') && p < r)
			p++;
		if (*p == '#' || *p == '\n')
			continue;
		q = targ;
		while (*p == *q && p < r)
			p++, q++;
		if (*q)
			continue;
		found = TRUE;
		q = data;
		s = data + size;
		while ((*p == ' ' || *p == '\t') && p < r)
			p++;
		/* 'loop' if there is an escaped '\n' */
	 loop:
		p = line;
		while ((*p == ' ' || *p == '\t') && p < r)
			p++;
		while (*p && q < s)
			*q++ = *p++;
		while (*q == ' ' || *q == '\t' || *q == '\n')
			q--;
		if (*q == '\\') {
			while (*q == ' ' || *q == '\t')
				q--;
			q++;
			*q++ = ' ';
			fgets(line, sizeof(line), sys_fd);
			goto loop;
		}
		*(++q) = '\0';
	}
	return (found);
}
