
#include "call_em.h"

extern struct defs *defs_base;
extern int parse_gettydefs(struct defs *, FILE *);

int debug;

void usage();
void show_termio(struct termios *);

/*
 * May be called directly or exec'ed by the '-c' option on the command
 * line of 'my_getty'.  This prints the interpretation of the
 * designated gettydefs file into ascii for approval.
 */

int
main(int argc, char *argv[])
{
	FILE *defs_file;
	struct defs here;
	char inbuf[512];
	char *p, *q, *fname;
	int once, offset;
	int i;

	debug = 0;
	i = 0;
	fname = DEFS_FILE;

	while (i < argc) {
		if (!(p = argv[i++]))
			break;
		if (*p == '-') {
			switch (*++p) {
			case 'd':
			case 'x':				  /* set the debug level */
				if (!*++p)
					p = argv[i++];
				debug = atoi(p);
				break;
			case '?':
				usage();
			}
		} else if (*p == '?')
			usage();
		else if ((access(p, R_OK)) == 0)
			fname = p;
	}

	strncpy(inbuf, fname, sizeof(inbuf));
	if ((defs_file = fopen(inbuf, "r")) == NULL) {
		printf("Unable to open \"%s\": %s\n",
				 inbuf, sys_errlist[errno]);
		return (-1);
	}

	/*
	 * This operation isn't very clean or imaginative but it is rarely
	 * used.
	 */
	once = 0;
	while (fgets(inbuf, sizeof(inbuf), defs_file) != NULL) {
		p = inbuf;
		while (*p == ' ' || *p == '\t')
			p++;
		if (*p == '#')
			continue;
		if (*p == '\n') {
			once++;
			continue;
		}
		while (*p == ' ' || *p == '\t')
			p++;
		if (once) {
			q = here.name;
			while (*p && *p != '#' && *p != ' ' && *p != '\t')
				*q++ = *p++;
			*q = '\0';
			offset = ftell(defs_file);
			rewind(defs_file);
			if (parse_gettydefs(&here, defs_file) < 0) {
				printf("Error reading defs file, %s\n",
						 sys_errlist[errno]);
				return (-1);
			}
			printf("\nCurrent entry: %s\nLocal 'termio':\n",
					 here.name);
			show_termio(&here.local);
			printf("\nRun 'termios':\n");
			show_termio(&here.run);
			printf("Prompt: %s\n", here.prompt);
			printf("Next entry: %s\n", here.next);
			fseek(defs_file, offset, 0);
		}
		/* 'once' is a flag to indicate a blank line when > 0 */
		once = 0;
	}
	return (0);
}

/*
 * Used by check_all() to translate termios flags into ascii for
 * appraisal.
 */
void
show_termio(struct termios *in_use)
{
	tcflag_t i, j;
	char *msg;

	msg = "";
	printf("c_iflag: 0x%.8lx ", in_use->c_iflag);
	for (i = 1; i <= IMAXBEL; i = i << 1) {
		if ((j = i & in_use->c_iflag) == 0)
			continue;
		switch (j) {
		case IGNBRK:
			msg = "IGNBRK";
			break;
		case BRKINT:
			msg = "BRKINT";
			break;
		case IGNPAR:
			msg = "IGNPAR";
			break;
		case PARMRK:
			msg = "PARMRK";
			break;
		case INPCK:
			msg = "INPCK";
			break;
		case ISTRIP:
			msg = "ISTRIP";
			break;
		case INLCR:
			msg = "INLCR";
			break;
		case IGNCR:
			msg = "IGNCR";
			break;
		case ICRNL:
			msg = "ICRNL";
			break;
		case IUCLC:
			msg = "IUCLC";
			break;
		case IXON:
			msg = "IXON";
			break;
		case IXANY:
			msg = "IXANY";
			break;
		case IXOFF:
			msg = "IXOFF";
			break;
		case IMAXBEL:
			msg = "IMAXBEL";
			break;
		}
		printf("%s ", msg);
	}
	printf("\n");
	printf("c_oflag: 0x%.8lx ", in_use->c_oflag);
	for (i = 1; i <= FF1; i = i << 1) {
		if (i <= OFDEL) {
			if ((j = i & in_use->c_oflag) == 0)
				continue;
			switch (j) {
			case OPOST:
				msg = "OPOST";
				break;
			case OLCUC:
				msg = "OLCUC";
				break;
			case ONLCR:
				msg = "ONLCR";
				break;
			case OCRNL:
				msg = "OCRNL";
				break;
			case ONOCR:
				msg = "ONOCR";
				break;
			case ONLRET:
				msg = "ONLRET";
				break;
			case OFILL:
				msg = "OFILL";
				break;
			case OFDEL:
				msg = "OFDEL";
				break;
			}
		} else if (i == NLDLY) {
			if ((i & in_use->c_oflag) == NL1)
				msg = "NL1";
			else
				msg = "NL0";
		} else if (i & CRDLY) {
			if ((j = in_use->c_oflag & CR3) == CR3)
				msg = "CR3";
			else if (j == CR2)
				msg = "CR2";
			else if (j == CR1)
				msg = "CR1";
			else
				msg = "CR0";
			i = 02000;
		} else if (i & TABDLY) {
			if ((j = in_use->c_oflag & TAB3) == TAB3)
				msg = "TAB3";
			else if (j == TAB2)
				msg = "TAB2";
			else if (j == TAB1)
				msg = "TAB1";
			else
				msg = "TAB0";
			i = 010000;
		} else if (i & BSDLY) {
			if ((j = in_use->c_oflag & BS1) == BS1)
				msg = "BS1";
			else
				msg = "BS0";
			i = 020000;
		} else if (i & VTDLY) {
			if ((j = in_use->c_oflag & VT1) == VT1)
				msg = "VT1";
			else
				msg = "VT0";
			i = 040000;
		} else if (i & FFDLY) {
			if ((j = in_use->c_oflag & FF1) == FF1)
				msg = "FF1";
			else
				msg = "FF0";
			i = 0100000;
		}
		printf("%s ", msg);
	}
	printf("\n");
	printf("c_cflag: 0x%.8lx ", in_use->c_cflag);
	i = in_use->c_cflag & (CBAUDEX | CBAUD);
	switch (i) {
	case B0:
		msg = "B0";
		break;
	case B50:
		msg = "B50";
		break;
	case B75:
		msg = "B75";
		break;
	case B110:
		msg = "B110";
		break;
	case B134:
		msg = "B134";
		break;
	case B150:
		msg = "B150";
		break;
	case B200:
		msg = "B200";
		break;
	case B300:
		msg = "B300";
		break;
	case B600:
		msg = "B600";
		break;
	case B1200:
		msg = "B1200";
		break;
	case B1800:
		msg = "B1800";
		break;
	case B2400:
		msg = "B2400";
		break;
	case B4800:
		msg = "B4800";
		break;
	case B9600:
		msg = "B9600";
		break;
	case B19200:
		msg = "B19200";
		break;
	case B38400:
		msg = "B38400";
		break;
	case B57600:
		msg = "B57600";
		break;
	case B115200:
		msg = "B115200";
		break;
	}
	printf("%s ", msg);
	if ((j = in_use->c_cflag & CSIZE) == CS5)
		msg = "CS5";
	else if (j == CS6)
		msg = "CS6";
	else if (j == CS7)
		msg = "CS7";
	else
		msg = "CS8";
	printf("%s ", msg);
	for (i = CSTOPB; i <= CLOCAL; i <<= 1) {
		if ((j = i & in_use->c_cflag) == 0)
			continue;
		switch (j) {
		case CSTOPB:
			msg = "CSTOPB";
			break;
		case CREAD:
			msg = "CREAD";
			break;
		case PARENB:
			msg = "PARENB";
			break;
		case PARODD:
			msg = "PARODD";
			break;
		case HUPCL:
			msg = "HUPCL";
			break;
		case CLOCAL:
			msg = "CLOCAL";
			break;
		}
		printf("%s ", msg);
	}
	if ((in_use->c_cflag & CIBAUD) == CIBAUD)
		printf("CIBAUD  ");
	if ((in_use->c_cflag & CRTSCTS) == CRTSCTS)
		printf("CRTSCTS ");
	printf("\n");
	printf("c_lflag: 0x%.8lx ", in_use->c_lflag);
	for (i = 1; i <= IEXTEN; i = i << 1) {
		if ((j = i & in_use->c_lflag) == 0)
			continue;
		switch (j) {
		case ISIG:
			msg = "ISIG";
			break;
		case ICANON:
			msg = "ICANON";
			break;
		case XCASE:
			msg = "XCASE";
			break;
		case ECHO:
			msg = "ECHO";
			break;
		case ECHOE:
			msg = "ECHOE";
			break;
		case ECHOK:
			msg = "ECHOK";
			break;
		case ECHONL:
			msg = "ECHONL";
			break;
		case NOFLSH:
			msg = "NOFLUSH";
			break;
		case TOSTOP:
			msg = "TOSTOP";
			break;
		case ECHOCTL:
			msg = "ECHOCTL";
			break;
		case ECHOPRT:
			msg = "ECHOPRT";
			break;
		case ECHOKE:
			msg = "ECHOKE";
			break;
		case FLUSHO:
			msg = "FLUSHO";
			break;
		case PENDIN:
			msg = "PENDIN";
			break;
		case IEXTEN:
			msg = "IEXTEN";
			break;
		}
		printf("%s ", msg);
	}
	printf("\n");
	printf("c_line:  0x%.8x ", in_use->c_line);
	switch (in_use->c_line) {
	case N_TTY:
		msg = "N_TTY";
		break;
	case N_SLIP:
		msg = "N_SLIP";
		break;
	case N_MOUSE:
		msg = "N_MOUSE";
		break;
	case N_PPP:
		msg = "N_PPP";
		break;
	}
	printf("%s\n", msg);
	printf("c_cc:\n");
	for (i = 0; i <= 6; i++) {
		cc_t ch, ch2;
		char msg2[8];
		char *p;
		int j, k;

		for (j = 0; j < 3; j++) {
			if (j == 0)
				k = i + j;
			else if (j == 1)
				k = i + 7;
			else
				k = i + 13;
			switch (k) {
			case VINTR:
				msg = "VINTR   ";
				break;
			case VQUIT:
				msg = "VQUIT   ";
				break;
			case VERASE:
				msg = "VERASE  ";
				break;
			case VKILL:
				msg = "VKILL   ";
				break;
			case VEOF:
				msg = "VEOF    ";
				break;
			case VTIME:
				msg = "VTIME   ";
				sprintf(msg2, "%2d", in_use->c_cc[i]);
				break;
			case VMIN:
				msg = "VMIN    ";
				sprintf(msg2, "%2d", in_use->c_cc[i]);
				break;
			case VSWTC:
				msg = "VSWTC   ";
				break;
			case VSTART:
				msg = "VSTART  ";
				break;
			case VSTOP:
				msg = "VSTOP   ";
				break;
			case VSUSP:
				msg = "VSUSP   ";
				break;
			case VEOL:
				msg = "VEOL    ";
				break;
			case VREPRINT:
				msg = "VREPRINT";
				break;
			case VDISCARD:
				msg = "VDISCARD";
				break;
			case VWERASE:
				msg = "VWERASE ";
				break;
			case VLNEXT:
				msg = "VLNEXT  ";
				break;
			case VEOL2:
				msg = "VEOL2   ";
				break;
			default:
				msg = "UNUSED  ";
				break;
			}
			ch = in_use->c_cc[k];
			ch2 = ch;
			if (k != VTIME && k != VMIN) {
				p = msg2;
				if (ch > 0x7f) {
					*p++ = 'M';
					*p++ = '-';
					ch -= 0x80;
				}
				if (ch < ' ') {
					*p++ = '^';
					ch += 0x40;
				} else if (ch == 0x7f) {
					*p++ = '^';
					ch = '?';
				}
				*p++ = ch;
				*p = '\0';
			}
			printf("   %s 0x%.2x = %s", msg, ch2, msg2);
			if (i == 6)
				break;
		}
		printf("\n");
	}
}

void
usage()
{
	fprintf(stdin, "check_getty: [-xN] gettydefs-type-file\n");
	exit(0);
}
