
#include "call_em.h"

extern int debug;
extern void show_termio(struct termios *);
struct defs *defs_base = NULL;

struct entry *process_flag(char *, char *);
void do_err(int, char *, char *, char *, char *, int);
int set_speed(int, enum how);
int read_defs(char *, char *);
int parse_gettydefs(struct defs *, FILE *);

struct entry {
	char name[10];
	union {
		tcflag_t flag;
		cc_t control;
	} def_flag;
	int type;
};

#define my_iflag 0
#define my_oflag 1
#define my_cflag 2
#define my_lflag 3
#define my_line  4
#define my_cc    5

/*
 * This array must be alphabetized according to the string entry. It is
 * used by parse_flag() to translate the ascii string into termios
 * flags.  The numeric third entry indicates to which field of the
 * termios structure this flag is found.
 */
struct entry flag_names[] =
{
	{"B0",       B0,       my_cflag},
	{"B110",     B110,     my_cflag},
	{"B1200",    B1200,    my_cflag},
	{"B134",     B134,     my_cflag},
	{"B150",     B150,     my_cflag},
	{"B1800",    B1800,    my_cflag},
	{"B19200",   B19200,   my_cflag},
	{"B200",     B200,     my_cflag},
	{"B2400",    B2400,    my_cflag},
	{"B300",     B300,     my_cflag},
	{"B38400",   B38400,   my_cflag},
	{"B4800",    B4800,    my_cflag},
	{"B50",      B50,      my_cflag},
	{"B600",     B600,     my_cflag},
	{"B75",      B75,      my_cflag},
	{"B9600",    B9600,    my_cflag},
	{"BRKINT",   BRKINT,   my_iflag},
	{"BS0",      BS0,      my_oflag},
	{"BS1",      BS1,      my_oflag},
	{"BSDLY",    BSDLY,    my_oflag},
	{"CIBAUD",   CIBAUD,   my_cflag},
	{"CLOCAL",   CLOCAL,   my_cflag},
	{"CR0",      CR0,      my_oflag},
	{"CR1",      CR1,      my_oflag},
	{"CR2",      CR2,      my_oflag},
	{"CR3",      CR3,      my_oflag},
	{"CREAD",    CREAD,    my_cflag},
	{"CRTSCTS",  CRTSCTS,  my_cflag},
	{"CS5",      CS5,      my_cflag},
	{"CS6",      CS6,      my_cflag},
	{"CS7",      CS7,      my_cflag},
	{"CS8",      CS8,      my_cflag},
	{"CSTOPB",   CSTOPB,   my_cflag},
	{"ECHO",     ECHO,     my_lflag},
	{"ECHOCTL",  ECHOCTL,  my_lflag},
	{"ECHOE",    ECHOE,    my_lflag},
	{"ECHOK",    ECHOK,    my_lflag},
	{"ECHOKE",   ECHOKE,   my_lflag},
	{"ECHONL",   ECHONL,   my_lflag},
	{"ECHOPRT",  ECHOPRT,  my_lflag},
	{"EXTA",     EXTA,     my_cflag},
	{"EXTB",     EXTB,     my_cflag},
	{"FF0",      FF0,      my_oflag},
	{"FF1",      FF1,      my_oflag},
	{"FFDLY",    FFDLY,    my_oflag},
	{"FLUSHO",   FLUSHO,   my_lflag},
	{"HUPCL",    HUPCL,    my_cflag},
	{"ICANON",   ICANON,   my_lflag},
	{"ICRNL",    ICRNL,    my_iflag},
	{"IEXTEN",   IEXTEN,   my_lflag},
	{"IGNBRK",   IGNBRK,   my_iflag},
	{"IGNPAR",   IGNPAR,   my_iflag},
	{"IMAXBEL",  IMAXBEL,  my_iflag},
	{"INLCR",    INLCR,    my_iflag},
	{"INPCK",    INPCK,    my_iflag},
	{"ISIG",     ISIG,     my_lflag},
	{"ISTRIP",   ISTRIP,   my_iflag},
	{"IUCLC",    IUCLC,    my_iflag},
	{"IXANY",    IXANY,    my_iflag},
	{"IXOFF",    IXOFF,    my_iflag},
	{"IXON",     IXON,     my_iflag},
	{"NL0",      NL0,      my_oflag},
	{"NL1",      NL1,      my_oflag},
	{"NLDLY",    NLDLY,    my_oflag},
	{"NOFLSH",   NOFLSH,   my_lflag},
	{"N_MOUSE",  N_MOUSE,  my_line},
	{"N_PPP",    N_PPP,    my_line},
	{"N_SLIP",   N_SLIP,   my_line},
	{"N_TTY",    N_TTY,    my_line},
	{"OCRNL",    OCRNL,    my_oflag},
	{"OFDEL",    OFDEL,    my_oflag},
	{"OFILL",    OFILL,    my_oflag},
	{"OLCUC",    OLCUC,    my_oflag},
	{"ONLCR",    ONLCR,    my_oflag},
	{"ONLRET",   ONLRET,   my_oflag},
	{"ONOCR",    ONOCR,    my_oflag},
	{"OPOST",    OPOST,    my_oflag},
	{"PARENB",   PARENB,   my_cflag},
	{"PARMRK",   PARMRK,   my_iflag},
	{"PARODD",   PARODD,   my_cflag},
	{"PENDIN",   PENDIN,   my_lflag},
	{"TAB0",     TAB0,     my_oflag},
	{"TAB1",     TAB1,     my_oflag},
	{"TAB2",     TAB2,     my_oflag},
	{"TAB3",     TAB3,     my_oflag},
	{"TOSTOP",   TOSTOP,   my_lflag},
	{"VDISCARD", VDISCARD, my_cc},
	{"VEOF",     VEOF,     my_cc},
	{"VEOL",     VEOL,     my_cc},
	{"VEOL2",    VEOL2,    my_cc},
	{"VERASE",   VERASE,   my_cc},
	{"VINTR",    VINTR,    my_cc},
	{"VKILL",    VKILL,    my_cc},
	{"VLNEXT",   VLNEXT,   my_cc},
	{"VMIN",     VMIN,     my_cc},
	{"VQUIT",    VQUIT,    my_cc},
	{"VREPRINT", VREPRINT, my_cc},
	{"VSTART",   VSTART,   my_cc},
	{"VSTOP",    VSTOP,    my_cc},
	{"VSUSP",    VSUSP,    my_cc},
	{"VSWTC",    VSWTC,    my_cc},
	{"VT0",      VT0,      my_oflag},
	{"VT1",      VT1,      my_oflag},
	{"VTDLY",    VTDLY,    my_oflag},
	{"VTIME",    VTIME,    my_cc},
	{"VWERASE",  VWERASE,  my_cc},
	{"XCASE",    XCASE,    my_lflag},
	{"XTABS",    XTABS,    my_oflag},
	{"", 0, 0},};

char *
say_speed(struct termios *getty)
{
	struct entry *ent;

	for (ent = flag_names; ent->name[0] != '\0'; ent++) {
		if ((ent->type == 2) &&
			 (ent->def_flag.flag == (getty->c_cflag & CBAUD)))
			return (ent->name);
	}
	return ("error");
}

/*
 * Selects the current termios from the defs_base array generated by
 * parse_gettydefs().  Increments the offset counter if called with
 * which == bump (local mode ).  Successive calls will step through the
 * various speeds listed in /etc/gettydefs.
 */
int
set_speed(int fd, enum how which)
{
	static cur_defs = 0;
	static num_defs = 0;
	struct termios tmp;
	char *msg;

	ioctl(fd, TCGETS, &tmp);
	switch (which) {
	case bump:
		if (++cur_defs > num_defs)
			cur_defs = 0;			  /* Fall through */
		fprintf(stderr, "bump: cur_defs %d, num_defs %d\n", cur_defs, num_defs);
	case local:
		tmp = (defs_base + cur_defs)->local;
		msg = say_speed(&tmp);
		break;
	case run:
		tmp = (defs_base + cur_defs)->run;
		msg = "run";
		break;
	case r_raw:
		tmp.c_cc[VMIN] = 0;
		tmp.c_cc[VTIME] = 0;
		tmp.c_lflag &= ~ICANON;
		msg = "raw";
		break;
	case cooked:
		tmp.c_cc[VMIN] = 1;
		tmp.c_cc[VTIME] = 0;
		tmp.c_lflag |= ICANON;
		msg = "cooked";
		break;
	case toloc:
		tmp.c_cflag |= CLOCAL;
		msg = "clocal on";
		break;
	case notloc:
		tmp.c_cflag &= ~CLOCAL;
		msg = "clocal off";
		break;
	case par:
		tmp.c_cflag |= PARENB;
		msg = "parity on";
		break;
	case nopar:
		tmp.c_cflag &= ~PARENB;
		msg = "parity off";
		break;
	case echo:
		tmp.c_lflag |= ECHO;
		msg = "echo";
		break;
	case noecho:
		tmp.c_lflag &= ~ECHO;
		msg = "no echo";
		break;
	case onlcr:
		tmp.c_oflag |= OPOST | ONLCR;
		msg = "onlcr";
		break;
	case restart:
		cur_defs = 0;
		num_defs = fd;
		return (0);
		break;
	default:
		msg = "";
		break;
	}
	if (ioctl(fd, TCSETSF, &tmp) != 0) {
		fprintf(stderr,
				  "set_speed: Unable to set termios: %s, %d, %d, %s\n",
				  sys_errlist[errno], fd, which, msg);
		return (-1);
	}
	if (debug > 2)
		fprintf(stderr, "Setting termios to %s\n", msg);
/* 	ioctl(fd, TCFLSH, 2); */
	sleep(1);
	return (cur_defs);
}

/*
 * Used during initialization to read the appropiate 'gettydefs' into
 * memory for use by the controling function.
 */
int
read_defs(char *initial, char *fname)
{
	FILE *defs_file;
	char *cur_name;
	struct defs *defs_curr;
	int i, ret, num_defs;
	char *p, *q;

	if (fname == NULL)
		fname = DEFS_FILE;
	if (debug > 2)
		fprintf(stderr, "\tLoading '%s' from '%s'.\n", initial, fname);
	if ((defs_file = fopen(fname, "r")) == NULL) {
		if (debug)
			fprintf(stderr, "Unable to open \"%s\": %s\n",
					  DEFS_FILE, sys_errlist[errno]);
		return (-1);
	}
	if (defs_base != NULL)
		free(defs_base);
	defs_base = (struct defs *) malloc(sizeof(struct defs));

	defs_curr = defs_base;
	num_defs = 0;
	cur_name = initial;
	while (1) {
		strncpy(defs_curr->name, cur_name, sizeof(defs_curr->name));
		rewind(defs_file);
		if ((ret = parse_gettydefs(defs_curr, defs_file)) == 0) {
			for (i = 0; i <= num_defs; i++) {
				p = (defs_base + i)->name;
				q = defs_curr->next;
				while (*p == *q && *p)
					p++, q++;
				if (!*p)
					goto done;
			}
			cur_name = defs_curr->next;
			num_defs++;
			defs_base = (struct defs *) realloc(defs_base,
										(sizeof(struct defs) * (num_defs + 1)));

			defs_curr = defs_base + num_defs;
		} else
			break;
	}
 done:
	fclose(defs_file);
	if (debug && num_defs == 0)
		fprintf(stderr, "Unable to parse gettydefs file for '%s'.\n",
				  initial);
	set_speed(num_defs, restart);
	return (num_defs ? 0 : -1);
}

/*
 * Used by read_defs() and check_all() to parse the 'gettydefs' file.
 * Each entry is read into the 'defs' structure pointed to by
 * *defs_curr.
 */
int
parse_gettydefs(struct defs *defs_curr, FILE * defs_file)
{
	char hold[512], inbuf[512];
	char *p, *q, *r, *s, *t, *u;
	int field, once, zap;
	struct termios *tmp;
	struct entry *ret;
	char *cur_entry;

	/* Second listing of definitions are from the red manual */
	cur_entry = defs_curr->name;
	tmp = &defs_curr->local;
	tmp->c_iflag        = 0;        /* input mode flags */
	tmp->c_oflag        = 0;        /* output mode flags */
	tmp->c_cflag        = 0;        /* control mode flags */
	tmp->c_lflag        = 0;        /* local mode flags */
	tmp->c_line         = N_TTY;    /* line discipline */
	tmp->c_cc[VINTR]    = CINTR;    /* (ETX ^C) (DEL ^@),  */
	tmp->c_cc[VQUIT]    = CQUIT;    /* (FS  ^|) (FS  ^|) */
	tmp->c_cc[VERASE]   = CERASE;   /* (BS  ^H) ('#') */
	tmp->c_cc[VKILL]    = CKILL;    /* (NAK ^U) ('@') */
	tmp->c_cc[VEOF]     = CEOF;     /* (EOT ^D) (EOT ^D) */
	tmp->c_cc[VTIME]    = 0;        /* No time delay */
	tmp->c_cc[VMIN]     = 1;        /* One character */
	tmp->c_cc[VSWTC]    = CSWITCH;  /* (0 sxtc) (EM  ^Z) */
	tmp->c_cc[VSTART]   = CSTART;   /* (DC1 ^Q) (DC1 ^Q) */
	tmp->c_cc[VSTOP]    = CSTOP;    /* (DC3 ^S) (DC3 ^S) */
	tmp->c_cc[VSUSP]    = CSUSP;    /* (SUB ^Z) (SUB ^Z) */
	tmp->c_cc[VEOL]     = CEOL;     /* (NUL \0) (NUL \0) */
	tmp->c_cc[VREPRINT] = CRPRNT;   /* (DC2 ^R) (DC2 ^R) */
	tmp->c_cc[VDISCARD] = CFLUSH;   /* (ST  ^O) (SI ^O) */
	tmp->c_cc[VWERASE]  = '\027';   /* (ETB ^W) (ETB ^W) */
	tmp->c_cc[VEOL2]    = CEOL2;    /* (NUL \0) undefined */
	tmp->c_cc[17]       = '\0';
	tmp->c_cc[18]       = '\0';
	memcpy(&defs_curr->run, &defs_curr->local, sizeof(struct termios));

	/*
	 * Shuffle through until we find a line which begins with the name
	 * we want.
	 */
	once = 0;
	cur_entry = defs_curr->name;
	while (1) {
		if (fgets(inbuf, sizeof(inbuf), defs_file) == NULL)
			return (-1);
		p = inbuf;
		while (*p == ' ' || *p == '\t')
			p++;
		if (*p == '#')
			continue;
		while (*p == ' ' || *p == '\t')
			p++;
		q = cur_entry;
		while (*p == *q)
			p++, q++;
		if (!*q)
			break;
	}
	while (*p != '#' && *p != '\n')
		p++;

	/*
	 * Now process the entry, one field at a time.
	 */
	field = 0;
	while (field < 4) {

		/*
		 * Read the next line and transfer the next field to the 'hold'
		 * buffer.
		 */
		p++;
		field++;
		q = hold;
		while (*p != '#' && *p != '\n') {
			while (*p && (*p != '\n' && *p != '#')) {
				/* Don't remove backslashes from the login prompt. */
				if (*p == '\\') {
					if (field == 3 && *(p + 1) == 'n') {
						*q++ = '\n';
						p++;
					}
					p++;
				} else
					*q++ = *p++;
			}
			if ((*p == '\0' || *p == '\n') && field != 4) {
				*q++ = ' ';
				if (fgets(inbuf, sizeof(inbuf), defs_file) == NULL)
					return (-2);
				p = inbuf;
				while (*p == ' ' || *p == '\t')
					p++;
			}
		}

		/*
		 * 'p' is our pointer into 'inbuf' so that we can continue
		 * reading in case the field separator is in the middle of a
		 * line.
		 */
		*q = '\0';
		r = hold;
		while (*r == ' ' || *r == '\t')
			r++;
		if (field == 1 || field == 2) {
			if (field == 1)
				tmp = &defs_curr->local;
			else
				tmp = &defs_curr->run;

			/*
			 * Scan for a 'SANE' entry.  If there is none, the various
			 * termios flags start as '0'.
			 */
			while (*r) {
				s = "SANE";
				while (*r == *s)
					r++, s++;

				/*
				 * If we found a match for 'SANE' in the hold buffer, 's'
				 * will be pointing at a '\0'.
				 */
				if (*s) {
					/* Move on to the next word. */
					while (*r != ' ' && *r != '\t')
						r++;
					while (*r == ' ' || *r == '\t')
						r++;
				} else {
					tmp->c_iflag = SANE_iflag;
					tmp->c_oflag = SANE_oflag;
					tmp->c_cflag = SANE_cflag;
					tmp->c_lflag = SANE_lflag;
					break;
				}
			}
			/* Start reading the input and process each flag, one by one. */
			t = hold;
			while (*t) {
				while (*t == ' ' || *t == '\t')
					t++;
				s = t;
				while (*t && *t != ' ' && *t != '\t' && *t != '\n')
					t++;
				if (*t == '\0' || *t == '\n')
					break;
				if (*s == '-') {
					zap = TRUE;
					s++;
				} else
					zap = FALSE;
				if ((ret = process_flag(s, t)) != NULL) {
					switch (ret->type) {
					case my_iflag:
						if (zap)
							tmp->c_iflag &= ~ret->def_flag.flag;
						else
							tmp->c_iflag |= ret->def_flag.flag;
						break;
					case my_oflag:
						if (zap)
							tmp->c_oflag &= ~ret->def_flag.flag;
						else {
							if (tmp->c_oflag & CRDLY &&
								 ret->def_flag.flag & CRDLY) {
								do_err(field, ret->name, "CRDLY", "c_oflag",
										 cur_entry, ret->type);
							} else if (tmp->c_oflag & TABDLY &&
										  ret->def_flag.flag & TABDLY) {
								do_err(field, ret->name, "TABDLY", "c_oflag",
										 cur_entry, ret->type);
							} else
								tmp->c_oflag |= ret->def_flag.flag;
						}
						break;
					case my_cflag:
						if (zap)
							tmp->c_cflag &= ~ret->def_flag.flag;
						else {
							if (tmp->c_cflag & CBAUD &&
								 ret->def_flag.flag & CBAUD) {
								do_err(field, ret->name, "CBAUD", "c_cflag",
										 cur_entry, ret->type);
							} else if (tmp->c_cflag & CSIZE &&
										  ret->def_flag.flag & CSIZE) {
								do_err(field, ret->name, "CSIZE", "c_cflag",
										 cur_entry, ret->type);
							} else
								tmp->c_cflag |= ret->def_flag.flag;
						}
						break;
					case my_lflag:
						if (zap)
							tmp->c_lflag &= ~ret->def_flag.flag;
						else
							tmp->c_lflag |= ret->def_flag.flag;
						break;
					}
				}
			}
			if (field == 1) {
				tmp->c_iflag |= PARMRK | INPCK;
				tmp->c_cflag |= PARENB | CLOCAL;
				tmp->c_lflag &= ~(ECHO | ICANON);
			}
		} else {
			if (field == 3) {
				t = defs_curr->prompt;
				u = t + sizeof(defs_curr->prompt) - 1;
			} else {					  /* field 4 */
				t = defs_curr->next;
				u = t + sizeof(defs_curr->next) - 1;
			}
			while (*r && *r != '#' && t < u)
				*t++ = *r++;
			*t = '\0';
		}
	}
	return (0);
}

/*
 * Put messages on the console when running with the -c option, when
 * running in background, write to the system error log.
 */
void
do_err(int field, char *name, char *group, char *flag, char *entry, int type)
{
	char form_1[] = "%s is a duplicate option to the %s pattern for %s in\n";
	char form_2[] = "\tthe %s field of %s.\n";

	if (getppid() == 1) {
		fprintf(stderr, form_1, name, group, flag);
		fprintf(stderr, form_2, type == 1 ? "local" : "run", entry);
	} else {
		printf(form_1, name, group, flag);
		printf(form_2, type == 1 ? "local" : "run", entry);
	}
}

/*
 * Used by parse_gettydefs() to search the 'flag_name[]' array in order
 * to translate ascii strings from the gettydefs file into termios flag
 * enumerations.
 */
struct entry *
process_flag(char *start, char *end)
{
	struct entry *bot, *mid, *top;
	register int i;
	char object[80];
	char *p, *q;
	int num;

	p = object;
	q = start;
	while (q < end)
		*p++ = *q++;
	*p = '\0';
	bot = flag_names;
	top = flag_names + (sizeof(flag_names) / sizeof(struct entry) - 1);

	num = (top - bot) / 2;
	mid = bot + num;

	while (1) {
		p = object;
		q = mid->name;
		if ((i = strcmp(p, q)) == 0)
			return (mid);
		if (i < 0) {
			top = mid;
			mid = bot + ((mid - bot) >> 1);
			if (mid == bot)
				break;
		} else if (i > 0) {
			bot = mid;
			mid = top - ((top - mid) >> 1);
			if (mid == top)
				break;
		}
	}
	return (NULL);
}
