#include "parse.h"

extern int set_defs(char *);
extern int check_locks(void);
extern int check_id(char *);
extern int set_devices(char *);
extern int set_locks();

extern struct defs *defs_base;

extern int debug;

int open_device(char *);
int check_time(char *);
char *read_response(int);
void display_char(unsigned char *, unsigned char);
int mpause(int m);

int
open_device(char *device)
{
	struct passwd *pw_ent;
	struct group *gr_ent;
	uid_t r_pid;
	__gid_t t_gid;
	int fd;
	char path[NAME_MAX];

	strcat(strcat(strcpy(path, DEVDIR), "/"), device);
	if (access(path, F_OK) < 0) {
		fprintf(stderr, "Can't access %s, %s", path, sys_errlist[errno]);
		return (-1);
	}
	/* Put the names of device and alternate in global reference */
	if ((set_devices(device)) < 0)
		return (-1);
	if ((set_locks(device)) < 0)
		return (-1);
	pw_ent = getpwnam(ROOT);
	r_pid = pw_ent->pw_uid;
	gr_ent = getgrnam(TTY);
	t_gid = gr_ent->gr_gid;

	/*
	 * Reset these because 'login' confiscates them for some dumb
	 * reason.
	 */
	chown(path, r_pid, t_gid);
	chmod(path, 0666);

	if ((fd = open(path, O_RDWR)) < 0) {
		fprintf(stderr, "Can't open %s, %s\n", device, sys_errlist[errno]);
		return (-1);
	} else if (debug > 2)
		fprintf(stderr, "\topen: Device '%s', fd = %d.\n", path, fd);
	return (fd);
}

/*
 * This function parses the TIME entries in the 'master' file. It
 * returns -1 on failure. A return value of 1, TRUE, means that the
 * current time is approved and a return value of 0 means that this in
 * not an approved time.
 */

int
check_time(char *list)
{
	long now;
	struct tm *tbuf;
	char *token[] =
	{
		"Su",
		"Mo",
		"Tu",
		"We",
		"Th",
		"Fr",
		"Sa",
		"Wk",
		"Any",
		"Never",
		NULL
	};
	char *p, *q, *r, *s, *t;
	char ch, buf[120];
	int day, start, end, rslt;
	int now_time;

	if (list == NULL)
		return (1);
	now = time(0L);
	tbuf = localtime(&now);

	start = end = rslt = -1;
	p = list;
	q = buf;
	while (*p && *p != ' ' && *p != '\n' && *p != '\t')
		*q++ = *p++;
	*q = '\0';
	p = buf;
	while (*p) {
		if (*p == '!') {
			p++;
			rslt = FALSE;
		} else
			rslt = TRUE;
		r = p;
		do {
			while (*r && *r != ',' && *r != ' ' && *r != '\t' && *r != '\n')
				r++;
			ch = *r;
			*r = '\0';
			for (day = 0; token[day]; day++) {
				s = p;
				t = token[day];
				while (*t && *s == *t)
					s++, t++;
				if (!*t) {
					start = atoi(s);
					while (*s++ != '-' && *s);
					end = atoi(s);
					while (*s >= '0' && *s <= '9')
						s++;
					if (*s && *s != ',' && *s != '\n' && *s != ' ' && *s != '\t')
						day = 10;
					break;
				}
			}
			if (day == 8)
				return (rslt);		  /* Any */
			else if (day == 9)
				return (!rslt);	  /* Never */
			else if (day > 9 || start > 2400 || end < 0) {
				*r = ch;
				fprintf(stderr, "check_time: error parsing %s\n", buf);
				return (-1);
			}
			now_time = tbuf->tm_hour * 100;
			now_time += tbuf->tm_min;
			if (day == 7 && tbuf->tm_wday > 0 && tbuf->tm_wday < 6)
				day = tbuf->tm_wday;
			if (!rslt)
				day = tbuf->tm_wday;
			if (start > end && now_time < end)
				day++;
			if (day > 6)
				day = 0;
			if (day == tbuf->tm_wday) {
				if (now_time >= start) {
					if (start > end || now_time < end) {
						return (rslt);
					}
				} else if (start > end && now_time < end) {
					return (rslt);
				} else {
					return (!rslt);
				}
			}
		} while (ch != '\n' && ch != ' ' && ch != '\t' && ch != ',' && ch);
		*r = ch;
		p = r;
		if (ch == ',')
			p++;
	}
	return (!rslt);
}

char *
read_response(int fd)
{
	static unsigned char buf[80];
	static unsigned char *p = buf;
	static unsigned char *q = buf + sizeof(buf);
	static unsigned char ch = '\0';
	int ret;

	while (((ret = read(fd, p, 1) > 0)) && p < q) {
		if (debug)
			display_char(NULL, *p);

*p &= 0x7f;
		/*
		 * I don't know why the input channel insists on sending a
		 * '<cr><nl><h>' sequence.  Anyhow, we dump it here.
		 */
		if (*p == 'h' && (ch == '\n' || ch == '\r')) {
			p = buf;
			ch = '\0';
			continue;
		} else if ((*p == '\n' || *p == '\r' || *p == '\0') && p != buf) {
			*p = '\0';
			p = buf;
			return (buf);
		}
		ch = *p;
		if (*p == '\b' && p > buf) {
			p--;
			write( fd, "\b", 1 );
		}
		else if (*p >= ' ' && *p <= '~')
			p++;
	}
	return (NULL);
}

/*
 * Put a character or string of characters on stderr prefixed with '^'
 * for control characters, with 'M-' for meta characters or with 'M-^'
 * for meta control characters.
 * 
 * msg == NULL, in_ch is considered a single character to be transformed
 * and written to stderr. msg != NULL, 'msg' is considered to be a NUL
 * terminated string to be transformed and writen to stderr. Neither
 * form sends a newline or carriage return.  These characters are
 * transformed along with all other control chars.
 */
void
display_char(unsigned char *msg, unsigned char in_ch)
{
	unsigned char buf[INBUF], *p, *q, *s, *r;
	unsigned char ch;

	p = buf;
	r = NULL;
	if (msg == NULL) {
		ch = in_ch;
		q = &ch;
		s = p + 1;
	} else {
		r = msg;
		while (*r)
			r++;
		r--;
		q = msg;
		s = buf + sizeof(buf) - 4;/* Leave room for one meta char plus
										   * NUL */
	}
	while (p < s && *q) {
		ch = *q++;
		if (ch > 0x7f) {
			*p++ = 'M';
			*p++ = '-';
			ch &= 0x7f;
		}
		if (ch < ' ') {
			*p++ = '^';
			*p++ = ch + 0x40;
		} else if (ch < 0x7f)
			*p++ = ch;
		else {
			*p++ = '^';
			*p++ = '?';
		}
	}
	if (r != NULL && *r != '\n') {
		*p++ = '\\';
		*p++ = 'c';
	}
	*p = '\0';
	fputs(buf, stderr);
}

#ifndef HZ
#include <linux/param.h>
#endif
#define CLOCK_TICK 1000/HZ		  /* HZ = 60 for unix-pc, 100 for linux */

int
mpause(int m)
{
	extern long times();
	long t1, t2, tarray[4];
	int t3;

	if (m <= 0)
		return (0);
	if ((t1 = times(tarray)) < 0)
		return (-1);
	while (1) {
		if ((t2 = times(tarray)) < 0)
			return (-1);
		t3 = ((int) (t2 - t1)) * CLOCK_TICK;
		if (t3 > m)
			return (t3);
	}
}										  /* mpause() */
