
#include "call_em.h"
#include <grp.h>
#include <unistd.h>

#define LOGFILE		"/var/log/cu"	 /* Where errors will be logged */

extern int set_defs(char *);
extern int check_locks(void);
extern int set_devices(char *);
extern int zap_locks();
extern int parse_sysfiles(char *, char *, int);
extern int set_speed(int, int);

extern struct defs *defs_base;

extern char pri_device[];
int debug;

void catch_int(int);
void catch_hup(int);
void catch_alarm(int);
void usage(void);

int
main(int argc, char *argv[])
{
	time_t now;
	char *p;
	char *system;
	int p_fd[2];

	int time_out;
	int i, fd, chld;
	struct termios term0, term1, term2;
	int ch1, ch2;

	time_out = 60;
	debug = 0;
	system = NULL;
	i = 0;
	while (i < argc) {
		if (!(p = argv[i++]))
			break;
		if (*p == '-') {
			switch (*++p) {
			case 't':				  /* set the timeout interval */
				if (!*++p)
					p = argv[i++];
				time_out = atoi(p);
				break;
			case 'd':				  /* Fall through */
			case 'x':				  /* set the debug level */
				if (!*++p)
					p = argv[i++];
				debug = atoi(p);
				break;
			case 'e':
				break;				  /* set even parity */
			case 'h':
				break;				  /* emulate local echo */
			case 'n':
				break;				  /* prompt for phone number */
			case 'o':
				break;				  /* set odd parity */
			case 's':
				if (!*++p)
					p = argv[i++];
				system = p;
				break;				  /* system */
			case 'l':
				break;				  /* line */
			}
		} else if (*p == '?')
			usage();
	}
	if (getppid() <= 1) {
		freopen(LOGFILE, "a", stderr);
		setvbuf(stderr, NULL, _IONBF, 0);
		time(&now);
		fprintf(stderr, "%.24s: %s, Startup\n", ctime(&now), pri_device);
	}
	if (check_locks() != 0) {
		fprintf(stderr, "%.24s: %s locks already set.\n",
				  ctime(&now), pri_device);
		exit(-1);
	}
	if( system != NULL ) {
		signal(SIGINT, catch_int);
		signal(SIGHUP, catch_hup);
		fd = parse_sysfiles(system, "cu", FALSE);
		if (fd > 0) {
			printf("Connected:\n");
			fflush(stdout);
			ioctl(0, TCGETS, &term0);
			ioctl(1, TCGETS, &term1);
			ioctl(2, TCGETS, &term2);
			if( pipe( p_fd ) < 0 ) {
				fprintf( stderr, "Unable to open pipe, %s\n", sys_errlist[errno] );
				goto getout;
			}
			if(( chld = fork()) < 0 ) {
				fprintf( stderr, "Unable to fork, %s\n", sys_errlist[errno] );
				close( p_fd[0] );
				close( p_fd[1] );
				goto getout;
			}
			/* read from [0] gets data written to [1] */
			else if( chld == 0 ) {
				/* Child process */
				exit;
			}
			else if( chld > 0 ) {
				/* Parent process */
			}
#if 0
			set_speed(fd, r_raw);
			set_speed(fd, noecho);
#endif
#if 0		/* This is under development */
			freopen( pri_device, "r", stdin );
			ioctl(0, TCGETS, &wrk_term);
			in_term = wrk_term;
			wrk_term.c_cc[VTIME] = 0;
			wrk_term.c_cc[VMIN] = 0;
			wrk_term.c_lflag &= ~(ICANON | ECHO);
			ioctl(0, TCSETS, &wrk_term);

			freopen( pri_device, "w", stdout );
			out_term = wrk_term;
			wrk_term.c_cc[VTIME] = 0;
			wrk_term.c_cc[VMIN] = 1;
			wrk_term.c_iflag &= ~ICRNL;
			wrk_term.c_iflag |= IGNCR;
			wrk_term.c_lflag &= ~(ICANON | ECHO);
			ioctl(1, TCSETS, &wrk_term);

			fprintf(stderr, "Read/Write loop ready.\n");
			do {
				while (read(0, &ch1, 1) > 0)
					write(fd, &ch1, 1);
				while (read(fd, &ch2, 1) > 0)
					write(1, &ch2, 1);
			} while ((char) ch1 != '~');
#endif
			close(fd);
			ioctl(0, TCSETS, &term0);
			ioctl(1, TCSETS, &term1);
			ioctl(3, TCSETS, &term2);
		}
	getout:
		zap_locks();
		if (fd > 0) {
			close(fd);
			exit(0);
		}
	} else
		usage();
	exit(1);
}

void
usage()
{
	fprintf(stdin,
			  "uucall: [-sspeed] [-lline] [-h] [-t##] [-d] [-o] [-e] [-n] telno\n");
	fprintf(stdin,
			  "uucall: [-h] [-d] [-o | -e] systemname\n");
	exit(0);
}

void
catch_hup(int sig)
{
	syslog(LOG_ERR, "Caught HANGUP, %d", sig);
	zap_locks();
	exit(-1);
}

void
catch_int(int sig)
{
	syslog(LOG_ERR, "Caught SIGINT, %d", sig);
	zap_locks();
	exit(-1);
}
