
#include "call_em.h"
#include <grp.h>
#include <unistd.h>

#define RECVFAX		"/usr/bin/netfax"	  /* Command line for receiving
													   * a fax */
#define LOGIN			"/bin/login"
#define LOGFILE		"/var/log/getty_em" /* Where errors will be logged */
#define CHECK_GETTY	"/usr/lib/uucp/check_em"

extern int set_speed(int, enum how);
extern int read_defs(char *, char *);
extern char *read_response(int);
extern void display_char(unsigned char *, unsigned char);
extern int set_locks();
extern int zap_locks();
extern int set_devices(char *);
extern int check_locks(void);
extern int check_id(char *);

extern struct defs *defs_base;

extern char pri_device[];
struct termios old, raw;
int debug;

void catch_int(int);
void catch_hup(int);
void catch_alarm(int);
int do_wait(char *, int);
int get_login(int, char *);
void usage(void);
int do_fax(int);
int get_type(char *);

int
main(int argc, char *argv[])
{
	time_t now;
	char *p, *fname;
	char getty_entry[30];
	int time_out;
	int two_way;
	int check_it;
	int i, j;

	fname = DEFS_FILE;
	two_way = FALSE;
	check_it = FALSE;
	time_out = 60;
	i = 0;

	while (i < argc) {
		if (!(p = argv[i++]))
			break;
		if (*p == '-') {
			switch (*++p) {
			case 'c':				  /* just check the 'gettydefs' file */
				check_it = TRUE;
				break;
			case 'r':				  /* two way communication on this line */
				two_way = TRUE;
				break;
			case 't':				  /* set the timeout interval */
				if (!*++p)
					p = argv[i++];
				time_out = atoi(p);
				break;
			case 'd':
			case 'x':				  /* set the debug level */
				if (!*++p)
					p = argv[i++];
				debug = atoi(p);
				break;
			}
		} else if (*p == '?')
			usage();
		else if ((set_devices(p)) == 0);
		else if ((access(p, X_OK)) == 0)
			fname = p;
		else
			strncpy(getty_entry, p, sizeof(getty_entry));
	}
	if (check_it) {
		printf("two_way: %s\n", two_way ? "TRUE" : "FALSE");
		printf("check_it: %s\n", check_it ? "TRUE" : "FALSE");
		printf("time_out: %d\n", time_out);
		printf("device: %s\n", pri_device);
		printf("entry requested: %s\n", getty_entry);
		printf("file name: %s\n\n", fname);
		execl(CHECK_GETTY, fname);
		exit(0);
	}
	if (getppid() <= 1) {
		freopen(LOGFILE, "a", stderr);
		setvbuf(stderr, NULL, _IONBF, 0);
		time(&now);
		fprintf(stderr, "\n%.24s: %s Startup\n", ctime(&now), pri_device);
	}
	if (getty_entry[0] == '\0')
		strcpy(getty_entry, "19200");	/* Set default */
	while (check_locks() != 0) {
		if (debug) {
			now = time(0L);
			fprintf(stderr, "%.24s: %s locks are set. Sleeping...\n",
					  ctime(&now), pri_device);
		}
		sleep(30);
	}
	j = 0;
	if ((i = read_defs(getty_entry, NULL) == 0))
		j = do_wait(pri_device, time_out);
	if (i < 0 || j < 0)
		sleep(60);					  /* Don't respawn too often */
	if (debug) {
		time(&now);
		fprintf(stderr, "%.24s: %s Exiting...\n",
				  ctime(&now), pri_device );
	}
	exit(0);
}

void
catch_hup(int sig)
{
	time_t now;

	time(&now);
	fprintf(stderr, "%.24s: %s Caught HANGUP\n", ctime(&now), pri_device);
	exit(-1);
}

void
catch_int(int sig)
{
	time_t now;

	time(&now);
	fprintf(stderr, "%.24s: %s Caught INTERRUPT\n", ctime(&now), pri_device);
	exit(-1);
}

void
catch_alarm(int sig)
{
	time_t now;

	time(&now);
	fprintf(stderr, "%.24s: %s Timed out.\n", ctime(&now), pri_device);
	exit(-1);
}

int
do_wait(char *device, int time_out)
{
	struct passwd *pw_ent;
	struct group *gr_ent;
	uid_t r_pid;
	__gid_t t_gid;
	int cur_getty;					  /* index into gettydefs array */
	int fd, ret, flags, back;
	int timer_on;
/* 	unsigned short ch1, ch2, ch3; */
	unsigned char ch1, ch2, ch3;
	time_t now;

	if (access(device, F_OK) < 0) {
		fprintf(stderr, "Can't access %s, %s\n", device, sys_errlist[errno]);
		return (-1);
	}
	pw_ent = getpwnam(ROOT);
	r_pid = pw_ent->pw_uid;
	gr_ent = getgrnam(TTY);
	t_gid = gr_ent->gr_gid;

	/*
	 * Reset these because 'login' confiscates them for some dumb
	 * reason.
	 */
	chown(device, r_pid, t_gid);
	chmod(device, 0666);

	while ((fd = open(device, O_RDWR)) < 0) {
		fprintf(stderr, "Can't open %s\n", device);
		sleep(20);
	}
	if ((cur_getty = set_speed(fd, local)) < 0) {
		fprintf(stderr, "Error setting initial speed, %s\n",
				  sys_errlist[errno]);
		close(fd);
		return (-1);
	}
	/* Toggle DTR just for laughs */
	flags = TIOCM_DTR | TIOCM_RTS;
	ioctl(fd, TIOCMBIC, &flags); /* Turn off flags */
	sleep(2);
	ioctl(fd, TIOCMBIS, &flags); /* Turn on flags */
ioctl(fd, TCFLSH, 2 );

	zap_locks();
	if (debug > 1)
		fprintf(stderr, "Waiting for CD ... \n");
	while (!flags & TIOCM_CD) {
		sleep(5);
		ioctl(fd, TIOCMGET, &flags);	 /* Read modem status */
	}
	signal(SIGINT, catch_int);
	timer_on = 0;
	ch2 = '\0';
	ch3 = '\0';
	back = 0;
	if (debug) {
		now = time(0L);
		fprintf(stderr, "%.24s: %s Waiting for chars ... \n",
				  ctime(&now), pri_device);
	}
errno = 0;
	while ((ret = read(fd, &ch1, sizeof(ch1))) == 1) {
		if (check_locks() < 0) {
			back = 0;
			break;
		}
		if (!timer_on) {
			timer_on++;
			if (debug) {
				now = time(0L);
				fprintf(stderr, "%.24s: %s Start timer, %d sec.\n",
						  ctime(&now), pri_device, time_out);
			}
			alarm(time_out);
		}
		if (debug)
			display_char(NULL, (char)ch1);

fprintf( stderr, "  0x%.4x\n", ch1);
		if ( ch1 == '\0' || ch1 == 0377) continue;
#if 0
		if ( ch2 == '\0' && ch3 == 0377) {
			if ((cur_getty = set_speed(fd, bump)) <= 0) {
				back = -1;
				break;
			}
		}
#endif
		else if (ch1 == 'X' && ch2 == 'A' && ch3 == 'F' ) {
/* 			do_fax(fd); */
			now = time(0L);
			fprintf( stderr, "do_wait %.24s\nFAX detected.\n", ctime(&now));
			back = 0;
			break;
		}
		else if (((ch1 == '\r') || (ch1 == '\n')) &&
					((ch2 == '\r') || (ch2 == '\n'))) {
			back = 1;
			break;
		}
		ch3 = ch2;
		ch2 = ch1;
	}
	if (back > 0) {
		if (debug)
			fputc('\n', stderr);
		signal(SIGHUP, catch_hup);
		back = get_login(fd, (defs_base + cur_getty)->prompt);
	}
	close(fd);
	return (back);
}

int
get_login(int fd, char *prompt)
{
	struct passwd *entry;
	static char query[] = "\nEnter Password: ";
	static char warning[] = "You must use lower case for login names.\n";
	static char bad_word[] = "Faulty password.\n";
	int caps;
	char *ret_word, *ep_word;
	char login[16];
	char buf[80];
	char *argv[10];
	char lsh[] = "login -pf ";
	char *p, *q, *path;
	time_t now, next;
	int count, len, ret, i;
	int index, stop, tries, good;

	now = time(0L);
	set_speed(fd, r_raw);
/* 	set_speed(fd, cooked); */
	set_speed(fd, nopar);
	tries = 4;						  /* Issue 'ogin:', 'sword:' prompts 4
										   * times then quit */
	ioctl(fd, TCFLSH, 2);
 retry:
	set_speed(fd, echo);
	len = strlen(prompt);
	ret_word = NULL;
	if (debug) {
		now = time(0L);
		fprintf(stderr, "%.24s: %s %start login.\n",
				  ctime(&now), pri_device, tries == 4 ? "S" : "Res");
	}
	ret = 0;
	entry = NULL;
	next = 0;
	for (count = 10; count;) {
		/* Don't issue prompt more often than 8 sec. */
		if (time(&now) >= next && ret_word == NULL ) {
			next = now + 8;
			count--;
			if( debug > 3 ) {
				fprintf(stderr, "Count %d\n", count);
				fprintf( stderr, "Sending: %s\n", prompt );
			}
			write(fd, prompt, len);
		}
		if ((ret_word = read_response(fd)) != NULL) {

			/*
			 * Make sure that the numbers on the next 2 lines correspond
			 * to the list in 'get_type'
			 */
			if ((ret = get_type(ret_word)) == 6 )
				return (do_fax(fd));

			/* We may have intruded on a call out routine. */
			if (ret == 0) {
				/* We stole their 'OK', give it back. */
				write(fd, "OK\n", 3);
				return (0);
			}
			if (ret == 7 || (check_locks() != 0))
				/* They are sending 'AT', get out now. */
				return (0);
			else if (ret < 8 )
				continue;

			/* 
			 * A return of 8 signifies the word wasn't in out list, so check
			 * the 'passwd' file for a match.
			 */
			setpwent();				  /* We reset before each try */
			if ((entry = getpwnam(ret_word)) != NULL)
				break;
			caps = 0;
			for( p = ret_word, q = p + 80; *p && p < q; p++ ) {
				if( *p >= 'A' && *p <= 'Z' ) caps++;
			}
			if( caps ) write( fd, warning, sizeof(warning));
			if (debug) {
				now = time(0L);
				fprintf(stderr, "%s Bad Login ret_word: '%s'\n",
						  pri_device, ret_word);
				ret_word = NULL;
			}
		}
	}
	if (count <= 0) {
		fprintf(stderr, "Attempt to read login failed 10 times.\n");
		return (0);
	}
	strncpy(login, ret_word, sizeof(login));
	if (debug) {
		now = time(0L);
		fprintf(stderr, "\n%.24s: %s Login '%s'\n",
				  ctime(&now), pri_device, login);
	}
	if ((strlen(entry->pw_shell) == 0) || (*entry->pw_passwd == '*')) {
		fprintf(stderr, "No shell provided for login '%s'\n", ret_word);
		return (0);
	}
	set_locks();
	/* Now get the password if one is required. */
	if (strlen(entry->pw_passwd) != 0) {
		len = strlen(query);
		if( debug > 2 )
				fprintf( stderr, "Sending: %s\n", query );
		write(fd, query, len);
		set_speed(fd, noecho);
		while ((ret_word = read_response(fd)) == NULL);
		ep_word = crypt(ret_word, entry->pw_passwd);
		if ((good = strncmp(entry->pw_passwd, ep_word,
								  strlen(entry->pw_passwd))) != 0) {
			fprintf(stderr, "\n%s Bad Password ret_word: '%s'\n",
					  pri_device, ret_word);
			len = strlen( bad_word );
			write( fd, bad_word, len);
			goto retry;
		}
		write( fd, "\n", 1 );
	}
	endpwent();
	if (debug) {
		now = time(0L);
		fprintf(stderr, "\n%.24s: %s Password '%s'\n",
				  ctime(&now), pri_device, ret_word);
	}
	set_speed(fd, run);
	if (entry->pw_gid == UUCP_GID) {
		if (setreuid(entry->pw_uid, UUCP_PID) < 0 && debug > 3)
			fprintf(stderr, "Cannot set user id to %d, %s\n",
					  entry->pw_uid, sys_errlist[errno]);
	} else {
		if (setgid(entry->pw_gid) < 0) {
			if (debug > 3)
				fprintf(stderr, "Cannot set group id to %d, %s\n",
						  entry->pw_gid, sys_errlist[errno]);
		}
		if (setpgid(0, 0) < 0) {
			if (debug > 3)
				fprintf(stderr, "Cannot set process group ID %d\n",
						  getpid());
		}
	}
	path = LOGIN;
	p = lsh;
	strncat(strncpy(buf, p, sizeof(buf) - 1),
			  entry->pw_name, sizeof(buf) - 1);

	if (debug > 3) {
		struct group *gr_ent;

		setgrent();
		gr_ent = getgrgid(entry->pw_gid);
		fprintf(stderr,
		  "Changing owner to '%s', pid %d, and group to '%s', gid %d\n",
		entry->pw_name, entry->pw_uid, gr_ent->gr_name, gr_ent->gr_gid);
		fprintf(stderr, "Exec'ing '%s', '%s'\n", path, buf);
	}
	chown(pri_device, entry->pw_uid, entry->pw_gid);

	/* Convert the string in buf to a series of args */
	p = buf;
	q = p;
	index = 0;
	stop = 0;
	while (!stop) {
		while (*q && *q != ' ')
			q++;
		if (!*q)
			stop = 1;
		*q++ = '\0';
		argv[index++] = p;
		p = q;
	}
	argv[index] = NULL;

	closelog();
	fclose(stderr);
	for (i = 0; i < fd; i++)
		close(i);
	for (i = 0; i < 3; i++)
		dup2(fd, i);
	if (fd >= 3) {
		close(fd);
		fd = 0;
	}
	execv(path, argv);
	fprintf(stderr, "exec failed to run %s", path);
	return (-1);
}

void
usage()
{
	fprintf(stdin, "getty_em: [-r] -t device entry\n");
	fprintf(stdin, "getty_em: -c gettydefs-type-file\n");
	fprintf(stdin, "getty_em: -q device\n");
	exit(0);
}

int
do_fax(int fd)
{
time_t now;
	if (set_locks() < 0)
		return (-1);
	/*
	 * We do a system() call here so that we will return and be able to
	 * zap our lock files.
	 */
now = time(0L);
fprintf( stderr, "do_fax %.24s\nFAX detected.\n", ctime(&now));
/* 	system(RECVFAX); */
	check_locks();					  /* if check_locks() finds our pid, it
										   * zaps it */
	return (0);
}

/* Parse the response code from modems. */
int
get_type(char *word)
{
	int ret, i;
	char *msg, *p, *q, *r;
	char buf[10];

	/* Do a check for numeric response codes */
	if (*word >= '0' && *word <= '9') {
		p = word;
		q = buf;
		r = q + sizeof(buf);
		while (*p && q < r)
			*q++ = *p++;
		*q = '\0';
		return (atoi(buf));
	}

	/*
	 * This list should correspond to most Hayes compatible modems. The
	 * 'for' loop which writes the login prompt in 'get_login' should be
	 * checked for logical consistency if any changes are made here.
	 */
	ret = 0;
	for (i = 0; i <= 8; i++) {
		switch (i) {
		case 0: msg = "OK";         break;
		case 1: msg = "CONNECT";    break;
		case 2: msg = "RING";       break;
		case 3: msg = "NO CARRIER"; break;
		case 4: msg = "ERROR";      break;
		case 5: msg = "DATA";       break;
		case 6: msg = "FAX";        break;
		case 7: msg = "AT";         break;
		default: msg = "";          break;
		}
		if (strcmp(word, msg) == 0)
			break;
	}
	return (i);
}

#if 0
/*
 * These are 'Result Codes' which may be returned depending upon
 * the setting of the 'X' and 'V' registers.
 */
/* Extended Result Codes enabled with X1, X2, X3 or X4 commsnds */
		case  5: msg = "CONNECT 1200";      break;
		case  6: msg = "NO DIALTONE";       break;
		case  7: msg = "BUSY";              break;
		case  8: msg = "NO ANSWER";         break;
		case  9: msg = "";                  break;
		case 10: msg = "CONNECT 2400";      break;
		case 11: msg = "CONNECT 4800";      break;
		case 12: msg = "CONNECT 9600";      break;
		case 16: msg = "CONNECT 12000";     break;
		case 13: msg = "CONNECT 14400";     break;
		case 14: msg = "CONNECT 19200";     break;  /* Enabled with X7 */
		case 15: msg = "CONNECT 7200";      break;
		case 17: msg = "CONNECT 16800";     break;
		case 18: msg = "";                  break;
/* Extended Result Codes ensbled with X5 (EC) */ 
		case  1: msg = "CONNECT 300";       break;
		case  5: msg = "CONNECT 1200/EC";   break;
		case 10: msg = "CONNECT 2400/EC";   break;
		case 11: msg = "CONNECT 4800/EC";   break;
		case 12: msg = "CONNECT 4800/EC";   break;
		case 13: msg = "CONNECT 14400/EC";  break;
		case 15: msg = "CONNECT 7200/EC";   break;
		case 16: msg = "CONNECT 12000/EC";  break;
		case 17: msg = "CONNECT 16800/EC";  break;
/* Extended Result Codes enabled with X7 */
		case 28: msg = "CONNECT 38400/EC";  break;
		case 30: msg = "CONNECT 57600/EC";  break;
		case 31: msg = "";                  break;
		case 32: msg = "CONNECT 76800/EC";  break;
		case 33: msg = "";                  break;
		case 34: msg = "CONNECT 115200/EC"; break;
/* Extended Result Codes ensbled with X6 (V.42) */ 
		case  1: msg = "CONNECT 300";       break;
		case  5: msg = "CONNECT 1200/V42";   break;
		case 10: msg = "CONNECT 2400/V42";   break;
		case 11: msg = "CONNECT 4800/V42";   break;
		case 12: msg = "CONNECT 4800/V42";   break;
		case 13: msg = "CONNECT 14400/V42";  break;
		case 15: msg = "CONNECT 7200/V42";   break;
		case 16: msg = "CONNECT 12000/V42";  break;
		case 17: msg = "CONNECT 16800/V42";  break;
/* Extended Result Codes enabled with X7 */
		case 28: msg = "CONNECT 38400/V42";  break;
		case 30: msg = "CONNECT 57600/V42";  break;
		case 31: msg = "";                  break;
		case 32: msg = "CONNECT 76800/EC";  break;
		case 33: msg = "";                  break;
		case 34: msg = "CONNECT 115200/EC"; break;
/* Extended Result Codes ensbled with X6 ( MNP ) */ 
		case  1: msg = "CONNECT 300";       break;
		case  5: msg = "CONNECT 1200/MNP";   break;
		case 10: msg = "CONNECT 2400/MNP";   break;
		case 11: msg = "CONNECT 4800/MNP";   break;
		case 12: msg = "CONNECT 4800/MNP";   break;
		case 13: msg = "CONNECT 14400/MNP";  break;
		case 15: msg = "CONNECT 7200/MNP";   break;
		case 16: msg = "CONNECT 12000/MNP";  break;
		case 17: msg = "CONNECT 16800/MNP";  break;
/* Extended Result Codes enabled with X7 */
		case 28: msg = "CONNECT 38400/MNP";  break;
		case 30: msg = "CONNECT 57600/MNP";  break;
		case 31: msg = "";                  break;
		case 32: msg = "CONNECT 76800/MNP";  break;
		case 33: msg = "";                  break;
		case 34: msg = "CONNECT 115200/MNP"; break;
#endif	/* if 0 */
