#include "pntrs.h"

extern int debug;

extern int pntr_init( struct p_pntrs *, int );

/*
 * Append the string pointed to by *strg to the area pointed to by
 * ch_pntr of struct *p_pntrx.  Realloc additional memory if needed.
 * Update the rest of the p_pntrx structure.
 */

int
add_on( struct p_pntrs *p_pntrx, char *strg )
{
	char *p, *q, *new_c;
	char **pp, **qq, **new_pp;
	int size, shift;
	int diff, blks, lines;

	/* If the p_pntrs struct hasn't been initialized, we can do it. */
	if( p_pntrx->ch_chunk == 0 ) {
		if( pntr_init( p_pntrx, 0 ) < 0 ) return( -1 );
	}
		
	/*
	 * Let's get out quick if we don't have anything to add.
	 */
	if (*strg == '\0')
		return (0);
	p = strg;
	q = p;
	lines = p_pntrx->pp_cnt;
	while (*q)
		if (*q++ == '\n')
			lines++;
	if (*(q - 1) != '\n') {
		lines++;
		q++;
	}
	size = q - p + p_pntrx->ch_size;
	shift = p_pntrx->ch_chunk;
	for (blks = p_pntrx->ch_blks;
		  (blks << shift) <= size + sizeof(char *); blks++);
	if (blks > p_pntrx->ch_blks) {
		new_c = (char *) realloc(p_pntrx->ch_pntr, blks << shift);
		if (new_c == NULL) {
			fprintf(stderr,
					"add_on: Unable to realloc memory for char buffer.\n");
			return (-1);
		}
		p_pntrx->ch_blks = blks;

			/*
			 * If the character buffer (ch_pntr) was moved, we have to
			 * change the pointer table (pp_pntr)
			 */
		if (p_pntrx->ch_pntr != new_c) {
			diff = new_c - p_pntrx->ch_pntr;
			for (pp = p_pntrx->pp_pntr, qq = pp + p_pntrx->pp_cnt;
				  pp <= qq; pp++)
				*pp += diff;
			*pp = NULL;
			p_pntrx->ch_pntr = new_c;
		}
	}
	for( blks = p_pntrx->pp_blks;
		  ( blks << 1 ) <= lines * sizeof(char **); blks++);
	if (blks > p_pntrx->pp_blks) {
		new_pp = (char **) realloc(p_pntrx->pp_pntr, blks * 128 );
		if (new_pp == NULL) {
			fprintf(stderr,
			  "add_on: Unable to realloc memory for buffer pointers.\n");
			return (-1);
		}
		p_pntrx->pp_pntr = new_pp;
		p_pntrx->pp_blks = blks;
	}
	p = p_pntrx->ch_pntr + p_pntrx->ch_size;
	q = strg;
	pp = p_pntrx->pp_pntr + p_pntrx->pp_cnt;
	*pp++ = p;
	while (*q) {
		if (*q == '\n') {
			*p++ = '\0';
			q++;
			if( *q ) *pp++ = p;
		}
		else *p++ = *q++;
	}
	*p = '\0';
	*pp = NULL;
	p_pntrx->ch_size = size;
	p_pntrx->pp_cnt = lines;

#ifdef DEBUG
	if( debug > 7 )
			fprintf(stderr, "
add_on:    ch_pntr  = 0x%07x, ch_size = 0x%07x, ch_blks = 0x%03x,
           pp_pntr  = 0x%07x,  pp_cnt = 0x%07x, pp_blks = 0x%03x,
           ch_chunk =       %03d\n",
				  (unsigned int)p_pntrx->ch_pntr,
				  p_pntrx->ch_size,
				  p_pntrx->ch_blks,
				  (unsigned int)p_pntrx->pp_pntr,
				  p_pntrx->pp_cnt,
				  p_pntrx->pp_blks,
				  p_pntrx->ch_chunk);
#endif /* DEBUG */
	return (0);
}
