#include "pntrs.h"
#include <sys/stat.h>
#include <errno.h>

extern int debug;


	/*
	 * This function allocates memory for a character file and allocates
	 * memory for a pointer file.  It then populates the pointer space
	 * with pointers to the beginning of each line in the character
	 * file.
	 */

int
mem_read( char *name, struct p_pntrs *p_pntrx )
{
	FILE *sfd;
	struct stat s_buf;
	char *p, *q, *c_pntr;
	char **pp, **p_pntr;
	off_t size;
	int lines, blks, blks_p, i;
	int shift;

	if( debug > 3 )
		fprintf( stderr, "mem_mgr: loading >%s<\n", name);

	if( stat(name, &s_buf ) < 0 ) {
		fprintf( stderr,
				  "mem_read: access denied to '%s'\n\tReason: %s\n",
				  name, sys_errlist[errno] );
		return( -1 );
	}
	size = s_buf.st_size;

	blks = size - 1;
	for( shift = 0; blks; shift++ ) blks >>= 1;
	if( shift < 7 ) shift = 7;

	blks = 1;
	if(( c_pntr = (char *) malloc(( blks << shift ))) == 0 ) {
		fprintf( stderr,
				  "mem_read: Unable to allocate memory for character buffer.\n");
		return (-1);
	}
	if(( sfd = fopen( name, "r" )) == NULL ) {
		fprintf(stderr,
				  "mem_read: Unable to open '%s'\n\t%s\n",
				  name, sys_errlist[errno] );
		return( -1 );
	}
	if(( i = fread( c_pntr, sizeof(char), size, sfd )) == 0 ) {
		fprintf( stderr,
				  "mem_read: Could read only %d bytes of %ld from '%s'.\n",
				  i, size, name );
		return( -1 );
	}
	p = c_pntr;
	q = p + size;
	lines = 0;

	while( p < q ) {
		if( *p == '\n' ) {
			lines++;
			*p = '\0';
		}
		p++;
	}
	*p = '\0';

	for( blks_p = 1; (blks_p << 1) <= (lines * sizeof(char **));
		  blks_p++ );

	if(( p_pntr = (char **) malloc( blks_p * 128 )) == 0 ) {
		fprintf(stderr,
				  "mem_read: Unable to allocate memory for pointer buffer.\n");
		free( c_pntr );
		fclose( sfd );
		return( -1 );
	}

	if( debug > 3 )
		fprintf(stderr, "mem_read: %d lines.\n", lines);

		/*
		 * Set up the pointer buffer with pointers to the beginning of
		 * each line.
		 */
	p = c_pntr;
	pp = p_pntr;
	while ( p < q) {
		*pp++ = p;
		while( *p++ );
	}
	*pp = NULL;

	fclose(sfd);
	p_pntrx->ch_pntr  = c_pntr;
	p_pntrx->ch_size  = size;
	p_pntrx->ch_blks  = blks;
	p_pntrx->ch_chunk = shift;
	p_pntrx->pp_pntr  = p_pntr;
	p_pntrx->pp_cnt   = lines;
	p_pntrx->pp_blks  = blks_p;

#ifdef DEBUG
	if( debug > 7 )
			fprintf(stderr, "
mem_read:  ch_pntr  = 0x%07x, ch_size = 0x%07x, ch_blks = 0x%03x,
           pp_pntr  = 0x%07x,  pp_cnt = 0x%07x, pp_blks = 0x%03x,
           ch_chunk =       %03d\n",
				  (unsigned int)p_pntrx->ch_pntr,
				  p_pntrx->ch_size,
				  p_pntrx->ch_blks,
				  (unsigned int)p_pntrx->pp_pntr,
				  p_pntrx->pp_cnt,
				  p_pntrx->pp_blks,
				  p_pntrx->ch_chunk);
#endif /* DEBUG */
	return (0);
}
