#include "pntrs.h"

extern int debug;

/*
 * Set initial sizes at a value inferred by 'select'.  If 'select'
 * is '0', or less than 127, the minimum size of chunks of 128 bytes
 * is assigned.  Values of 'select' greater than 127 is assumed to
 * the number of bytes desired per chunk.  This value is converted
 * to the next higher power of 2.  Values from 7 to 126 are taken
 * to be the size of chunks as powers of 2.
 */

int
pntr_init( struct p_pntrs *p_pntrx, int select )
{
	char *new_c;
	char **new_pp;
	int size, shift, blks;

	if ( select > 127 ) {
		blks = select - 1;
		for( shift = 0; blks; shift++ ) blks >>= 1;
	}						
	else if( select > 7 ) shift = select;
	else shift = 7;
	size = 1 << shift;

	if ((new_c =
		  (char *) malloc( size )) == NULL ) {
			/* Send error messages from calling routine. */
		return (-1);
	}
	if ((new_pp =
		  (char **) malloc( 128 )) == NULL ) {
			/* Send error messages from calling routine. */
		free(new_c);
		return (-1);
	}
	p_pntrx->ch_pntr = new_c;
	*p_pntrx->ch_pntr = '\0';
	p_pntrx->ch_size = 0;
	p_pntrx->ch_chunk = shift;
	p_pntrx->ch_blks = 1;
	p_pntrx->pp_pntr = new_pp;
	*p_pntrx->pp_pntr = NULL;
	p_pntrx->pp_cnt = 0;
	p_pntrx->pp_blks = 1;

#ifdef DEBUG
	if (debug > 7) {
		fprintf(stderr, "
init_pntr: ch_pntr  = 0x%07x, ch_size = 0x%07x, ch_blks = 0x%03x,
           pp_pntr  = 0x%07x,  pp_cnt = 0x%07x, pp_blks = 0x%03x,
           ch_chunk =       %03d\n",
				  (unsigned int)p_pntrx->ch_pntr,
				  p_pntrx->ch_size,
				  p_pntrx->ch_blks,
				  (unsigned int)p_pntrx->pp_pntr,
				  p_pntrx->pp_cnt,
				  p_pntrx->pp_blks,
				  p_pntrx->ch_chunk);
	}
#else
	if( debug > 4 )
			fprintf( stderr, "pntr_init: Okay.\n" );
#endif /* DEBUG */
	return (0);
}
