#include "pntrs.h"
#include "call_em.h"
extern int debug;

struct p_pntrs NULLpntrs =
{
	NULL, 0, 0, 0, NULL, 0, 0
};

int pntr_init( struct p_pntrs *, int );
int add_on( struct p_pntrs *, char * );
int mem_read( char *, struct p_pntrs * );
char **bin_search( char **, char *, int );
int pcompar( char **, char ** );
int compar( char *, char * );
int insert( struct p_pntrs *, char * );

/*
 * Set initial sizes at a value inferred by 'select'.  If 'select'
 * is '0', or less than 127, the minimum size of chunks of 128 bytes
 * is assigned.  Values of 'select' greater than 127 is assumed to
 * the number of bytes desired per chunk.  This value is converted
 * to the next higher power of 2.  Values from 7 to 126 are taken
 * to be the size of chunks as powers of 2.
 */

int
pntr_init( struct p_pntrs *p_pntrx, int select )
{
	char *new_c;
	char **new_pp;
	int size, shift, blks;

	if ( select > 127 ) {
		blks = select - 1;
		for( shift = 0; blks; shift++ ) blks >>= 1;
	}						
	else if( select > 7 ) shift = select;
	else shift = 7;
	size = 1 << shift;

	if ((new_c =
		  (char *) malloc( size )) == NULL ) {
			/* Send error messages from calling routine. */
		return (-1);
	}
	if ((new_pp =
		  (char **) malloc( 128 )) == NULL ) {
			/* Send error messages from calling routine. */
		free(new_c);
		return (-1);
	}
	p_pntrx->ch_pntr = new_c;
	*p_pntrx->ch_pntr = '\0';
	p_pntrx->ch_size = 0;
	p_pntrx->ch_chunk = shift;
	p_pntrx->ch_blks = 1;
	p_pntrx->pp_pntr = new_pp;
	*p_pntrx->pp_pntr = NULL;
	p_pntrx->pp_cnt = 0;
	p_pntrx->pp_blks = 1;

#ifdef DEBUG
	if (debug > 7) {
		fprintf(stderr, "
init_pntr: ch_pntr  = 0x%07x, ch_size = 0x%07x, ch_blks = 0x%03x,
           pp_pntr  = 0x%07x,  pp_cnt = 0x%07x, pp_blks = 0x%03x,
           ch_chunk =       %03d\n",
				  (unsigned int)p_pntrx->ch_pntr,
				  p_pntrx->ch_size,
				  p_pntrx->ch_blks,
				  (unsigned int)p_pntrx->pp_pntr,
				  p_pntrx->pp_cnt,
				  p_pntrx->pp_blks,
				  p_pntrx->ch_chunk);
	}
#else
	if( debug > 4 )
			fprintf( stderr, "pntr_init: Okay.\n" );
#endif /* DEBUG */
	return (0);
}

/*
 * Append the string pointed to by *strg to the area pointed to by
 * ch_pntr of struct *p_pntrx.  Realloc additional memory if needed.
 * Update the rest of the p_pntrx structure.
 */

int
add_on( struct p_pntrs *p_pntrx, char *strg )
{
	char *p, *q, *new_c;
	char **pp, **qq, **new_pp;
	int size, shift;
	int diff, blks, lines;

	/* If the p_pntrs struct hasn't been initialized, we can do it. */
	if( p_pntrx->ch_chunk == 0 ) {
		if( pntr_init( p_pntrx, 0 ) < 0 ) return( -1 );
	}
		
	/*
	 * Let's get out quick if we don't have anything to add.
	 */
	if (*strg == '\0')
		return (0);
	p = strg;
	q = p;
	lines = p_pntrx->pp_cnt;
	while (*q)
		if (*q++ == '\n')
			lines++;
	if (*(q - 1) != '\n') {
		lines++;
		q++;
	}
	size = q - p + p_pntrx->ch_size;
	shift = p_pntrx->ch_chunk;
	for (blks = p_pntrx->ch_blks;
		  (blks << shift) <= size + sizeof(char *); blks++);
	if (blks > p_pntrx->ch_blks) {
		new_c = (char *) realloc(p_pntrx->ch_pntr, blks << shift);
		if (new_c == NULL) {
			fprintf(stderr,
					"add_on: Unable to realloc memory for char buffer.\n");
			return (-1);
		}
		p_pntrx->ch_blks = blks;

			/*
			 * If the character buffer (ch_pntr) was moved, we have to
			 * change the pointer table (pp_pntr)
			 */
		if (p_pntrx->ch_pntr != new_c) {
			diff = new_c - p_pntrx->ch_pntr;
			for (pp = p_pntrx->pp_pntr, qq = pp + p_pntrx->pp_cnt;
				  pp <= qq; pp++)
				*pp += diff;
			*pp = NULL;
			p_pntrx->ch_pntr = new_c;
		}
	}
	for( blks = p_pntrx->pp_blks;
		  ( blks << 1 ) <= lines * sizeof(char **); blks++);
	if (blks > p_pntrx->pp_blks) {
		new_pp = (char **) realloc(p_pntrx->pp_pntr, blks * 128 );
		if (new_pp == NULL) {
			fprintf(stderr,
			  "add_on: Unable to realloc memory for buffer pointers.\n");
			return (-1);
		}
		p_pntrx->pp_pntr = new_pp;
		p_pntrx->pp_blks = blks;
	}
	p = p_pntrx->ch_pntr + p_pntrx->ch_size;
	q = strg;
	pp = p_pntrx->pp_pntr + p_pntrx->pp_cnt;
	*pp++ = p;
	while (*q) {
		if (*q == '\n') {
			*p++ = '\0';
			q++;
			if( *q ) *pp++ = p;
		}
		else *p++ = *q++;
	}
	*p = '\0';
	*pp = NULL;
	p_pntrx->ch_size = size;
	p_pntrx->pp_cnt = lines;

#ifdef DEBUG
	if( debug > 7 )
			fprintf(stderr, "
add_on:    ch_pntr  = 0x%07x, ch_size = 0x%07x, ch_blks = 0x%03x,
           pp_pntr  = 0x%07x,  pp_cnt = 0x%07x, pp_blks = 0x%03x,
           ch_chunk =       %03d\n",
				  (unsigned int)p_pntrx->ch_pntr,
				  p_pntrx->ch_size,
				  p_pntrx->ch_blks,
				  (unsigned int)p_pntrx->pp_pntr,
				  p_pntrx->pp_cnt,
				  p_pntrx->pp_blks,
				  p_pntrx->ch_chunk);
#endif /* DEBUG */
	return (0);
}

	/*
	 * This function allocates memory for a character file and allocates
	 * memory for a pointer file.  It then populates the pointer space
	 * with pointers to the beginning of each line in the character
	 * file.
	 */

int
mem_read( char *name, struct p_pntrs *p_pntrx )
{
	FILE *sfd;
	struct stat s_buf;
	char *p, *q, *c_pntr;
	char **pp, **p_pntr;
	off_t size;
	int lines, blks, blks_p, i;
	int shift;

	if( debug > 3 )
		fprintf( stderr, "mem_mgr: loading >%s<\n", name);

	if( stat(name, &s_buf ) < 0 ) {
		fprintf( stderr,
				  "mem_read: access denied to '%s'\n\tReason: %s\n",
				  name, sys_errlist[errno] );
		return( -1 );
	}
	size = s_buf.st_size;

	blks = size - 1;
	for( shift = 0; blks; shift++ ) blks >>= 1;
	if( shift < 7 ) shift = 7;

	blks = 1;
	if(( c_pntr = (char *) malloc(( blks << shift ))) == 0 ) {
		fprintf( stderr,
				  "mem_read: Unable to allocate memory for character buffer.\n");
		return (-1);
	}
	if(( sfd = fopen( name, "r" )) == NULL ) {
		fprintf(stderr,
				  "mem_read: Unable to open '%s'\n\t%s\n",
				  name, sys_errlist[errno] );
		return( -1 );
	}
	if(( i = fread( c_pntr, sizeof(char), size, sfd )) == 0 ) {
		fprintf( stderr,
				  "mem_read: Could read only %d bytes of %ld from '%s'.\n",
				  i, size, name );
		return( -1 );
	}
	p = c_pntr;
	q = p + size;
	lines = 0;

	while( p < q ) {
		if( *p == '\n' ) {
			lines++;
			*p = '\0';
		}
		p++;
	}
	*p = '\0';

	for( blks_p = 1; (blks_p << 1) <= (lines * sizeof(char **));
		  blks_p++ );

	if(( p_pntr = (char **) malloc( blks_p * 128 )) == 0 ) {
		fprintf(stderr,
				  "mem_read: Unable to allocate memory for pointer buffer.\n");
		free( c_pntr );
		fclose( sfd );
		return( -1 );
	}

	if( debug > 3 )
		fprintf(stderr, "mem_read: %d lines.\n", lines);

		/*
		 * Set up the pointer buffer with pointers to the beginning of
		 * each line.
		 */
	p = c_pntr;
	pp = p_pntr;
	while ( p < q) {
		*pp++ = p;
		while( *p++ );
	}
	*pp = NULL;

	fclose(sfd);
	p_pntrx->ch_pntr  = c_pntr;
	p_pntrx->ch_size  = size;
	p_pntrx->ch_blks  = blks;
	p_pntrx->ch_chunk = shift;
	p_pntrx->pp_pntr  = p_pntr;
	p_pntrx->pp_cnt   = lines;
	p_pntrx->pp_blks  = blks_p;

#ifdef DEBUG
	if( debug > 7 )
			fprintf(stderr, "
mem_read:  ch_pntr  = 0x%07x, ch_size = 0x%07x, ch_blks = 0x%03x,
           pp_pntr  = 0x%07x,  pp_cnt = 0x%07x, pp_blks = 0x%03x,
           ch_chunk =       %03d\n",
				  (unsigned int)p_pntrx->ch_pntr,
				  p_pntrx->ch_size,
				  p_pntrx->ch_blks,
				  (unsigned int)p_pntrx->pp_pntr,
				  p_pntrx->pp_cnt,
				  p_pntrx->pp_blks,
				  p_pntrx->ch_chunk);
#endif /* DEBUG */
	return (0);
}

#if 0		/* This isn't used by 'pppcall' */
char **
bin_search( char **p_pntr, char *name, int insert )
{
	register int i;
	char object[INBUF];
	char *p, *q;
	char **low, **mid, **top;
	int found;

	p = object;
	q = name;
	while( *q && *q != ' ' && *q != '\t' && *q != '\n' )
		*p++ = *q++;
	*p = '\0';

	low = p_pntr;
	for( top = low; *top != NULL; top++ );
	top--;
	mid = low + (( top - low ) >> 1 );

#ifdef DEBUG
	if( debug > 3 ) {
		fprintf( stderr, "bin_search: >%s<\n", name );
		fprintf( stderr, "\tlow 0x%07x, mid 0x%07x, top 0x%07x\n",
				  (int)low, (int)mid, (int)top );
	}
#endif	/* DEBUG */

	found = 0;
	while( 1 ) {
		p = object;
		q = *mid;
		if( debug > 5 ) {
			while (*q != ' ' && *q != '\n' && *q)
				q++;
			i = q - *mid;
			q = *mid;
			fprintf( stderr, "    object: >%.*s<\n", i, q );
		}
		if(( i = compar(p, q)) == 0 ) {
			found = 1;
			break;
		}
		if( i < 0 ) {
			top = mid;
			mid = low + ((mid - low) >> 1 );
			if( mid == low )
				break;
		} else if( i > 0 ) {
			low = mid;
			mid = top - ((top - mid) >> 1 );
			if( mid == top )
				break;
		}
	}
	if( compar( object, *mid ) == 0 )
			found = 1;
	return(( insert ^ found ) ? mid : (char **)0 );
}

int
pcompar( char **a, char **b )
{
	return( compar( *a, *b ));
}

int
compar( char *a, char *b )
{
	char *p, *q, *r, *s;
	char x, y;
	int i;

	if( debug > 4 )
			fprintf( stderr, "compar: >%s<\n        >%s<\n", a, b );
	if(( a == NULL ) || ( b == NULL )) {
		fprintf( stderr,
			  "compar: presented with a NULL pointer. a = 0x%07x, b = 0x%07x\n",
			  (int)a, (int)b );
		fprintf( stderr,
				  "compar: a = %s\n        b = %s\n",
				  a, b );
		return(0);
	}
	for( p = a;
		  *p != '\0' && *p != ' ' && *p != '\t' && *p != '\n'; p++ );
	x = *p;
	*p = '\0';
	for( q = b;
		  *q != '\0' && *q != ' ' && *q != '\t' && *q != '\n'; q++ );
	y = *q;
	*q = '\0';
	for( r = a, s = b; *r == *s && *r && *s; r++, s++ );
	i = (int) (*r - *s);
	*p = x;
	*q = y;
	return( i );
}

int
insert( struct p_pntrs *pntrs, char *data )
{
	char **pp, **qq, **rr;
	char *p;
	int position;

	if( debug > 2 )
		fprintf(stderr, "insert: data = >%s<.\n", data);

	/* The trivial condition for the first data to be inserted. */
	if( pntrs->pp_cnt == 0 ) {
		if( add_on( pntrs, data ) < 0 )
			return( -1 );
		return( 1 );
	}
	/* Another trivial condition with only one entry. */
	if (pntrs->pp_cnt == 1) {
		if( compar( *(pntrs->pp_pntr), data ) == 0 ) pp = NULL;
		else pp = pntrs->pp_pntr + 1;
	}
	/* No longer trivial, call 'bin_search()' with the 'insert' flag set. */
	else pp = bin_search( pntrs->pp_pntr, data, 1 );

	/* pp will point to the location the pointer should be inserted */
	if ( pp != NULL ) {
		position = pp - pntrs->pp_pntr;
		if( add_on( pntrs, data ) < 0 )
			return( -1 );
		/* pntrs->pp_pntr may get changed by add_on() if realloced */
		pp = pntrs->pp_pntr + position;
		qq = pntrs->pp_pntr + ( pntrs->pp_cnt - 1 );
		rr = qq - 1;
		if( rr < pntrs->pp_pntr ) rr = pntrs->pp_pntr;
		if( compar( *rr, *qq ) > 0 ) {
			while( pp > pntrs->pp_pntr && compar( *pp, data ) > 0 )
				pp--;
			while( pp <= rr && compar( *pp, data ) < 0 )
				pp++;
			p = *qq;
			while( rr >= pp ) {
				*qq = *rr;
				qq--; rr--;
			}
			*pp = p;
		}
		return( 1 );
	}
	return( 0 );
}
#endif	/* if 0 */
