#ifndef Pntr_h
#define Pntr_h

#include <stdio.h>
#include <malloc.h>
#include <string.h>
#include <sys/dirent.h>

#define INBUF			240
#define SHRT_BUF		64

struct p_pntrs {
	char *ch_pntr;
	int ch_size;
	int ch_chunk;
	int ch_blks;
	char **pp_pntr;
	int pp_cnt;
	int pp_blks;
};

#define ppFREE(p)	\
	if ( p.ch_pntr != NULL ) free(p.ch_pntr);	\
	if ( p.pp_pntr != NULL ) free(p.pp_pntr);	\
	p.ch_pntr  = NULL; \
	p.ch_size  = 0;    \
   p.ch_chunk = 0;    \
	p.ch_blks  = 0;    \
	p.pp_pntr  = NULL; \
	p.pp_cnt   = 0;    \
	p.pp_blks  = 0;

#endif	/* Pntr_h */
	

/* 
 * These are the global definitions which should be in an "include"
 * file:
 *    struct p_pntrs {
 *       char *ch_pntr;  The base pointer for the text data.
 *       int ch_size;    The number of bytes of data read in.
 *			int chunk;      The size of memory blocks.
 *       int ch_blks;    The number of blocks allocated.
 *       char **pp_pntr; The base pointer to the char * data.
 *       int pp_cnt;     The number of pointer entries.
 *       int pp_blks;    The number of blocks allocated for pointers.
 *    };
 *
 *    #define ppFREE(p)	\
 *       if ( p.ch_pntr != NULL ) free(p.ch_pntr);	\
 *       if ( p.pp_pntr != NULL ) free(p.pp_pntr);	\
 *       p.ch_pntr  = NULL; \
 *       p.ch_size  = 0; 	 \
 *       p.ch.chunk = 0;    \
 *       p.ch_blks  = 0;	 \
 *       p.pp_pntr  = NULL; \
 *       p.pp_cnt   = 0;	 \
 *       p.pp_blks  = 0;
 *
 * The idea is that an area of memory is allocated for holding text data for
 * processing.  The beginning of this area is referenced by ch_pntr.  As each
 * line is read in, a char pointer to the beginning of the line is stored in
 * another area of memory referenced by pp_pntr.
 *
 * Whenever an activity will overflow the allocated memory, the area is
 * realloc'ed.  Since realloc may move the data in the area if a contiguous
 * area is not available.  When realloc does move the data, the pointers
 * in the pp_pntr area are revised to point to the same data it did before
 * the move.
 *
 * 'pntr_init()' takes a pointer to a 'p_pntrs' struct and the size of the
 *               blocks to be allocated.  This number should be a power of 2.
 *               The default value is '7' ( 128 bytes ).  Values greater than
 *               127 will be assumed to reference the number of bytes desired.
 *               The number will be converted to the next smaller power of 2.
 *
 * 'add_on()'    takes a pointer to a 'p_pntrs' and a char pointer which
 *               points to a line of text which can be either terminated
 *               with a '\0' or a '\n'.  Data is converted to '\n' term-
 *               ated strings.
 *
 * 'mem_read()'  takes a pointer the path name of a file to be read into
 *               memory.  The file is opened and read into a newly allocated
 *               area of memory with pointers to the beginning of each line
 *               placed in another area of memory.  The information for
 *               referencing this data in the p_pntrs struct pointed to by
 *               the second argument.  The value for ch_chunks will be 7
 *               ( 128 bytes ) unless previously initialized.  Any memory
 *               previously allocated to this pointer will be feee()'d.
 *
 *  The chunks size for allocated the pointers area is always 128 or 16 lines
 *  of text.
 *
 * 'bin_search()' This takes a pointer to an array of pointers to char such as
 *               'pp_pntr' and a pointer to the name to be found.  It returns
 *               NULL if name was not found or a **char if the name was found.
 *               If the 'insert' flag = 1, a pointer to where the the item
 *               should be inserted is returned.  This feature is used by the
 *               'insert()' routine.
 *
 * 'pcompar()'   convert **char arguments to *char for use by 'compar()'.
 *               To be used when sorting data referenced with a 'p_pnts'
 *               structure.  Usage:
 *
 *               struct p_pntrs my_data;
 *
 *		           qsort((char **) my_data.pp_pntr, my_data.pp_cnt,
 *                     (char **) sizeof(char **), (int) pcompar);
 *
 * 'compar()'    A version of strcmp() which uses spaces, newlines and tabs
 *               as delimiters as well as '\0'. *a and *b must point to
 *               strings in the data space.  References to static space will
 *               cause  a segmentation fault.
 *
 * 'insert()'    This routine adds 'data' onto the end of a 'pntr->ch_pntr'
 *               area using 'add_on()'.  It then inserts the pointer to
 *               'data' into the previously sorted array of pointers,
 *               'pntrs->pp_pntr'. If another identical entry exists, the
 *               pointer is rejected.  A -1 is returned for error, 0 if the
 *               name has been seen before and +1 if this is the first time
 *               the name was seen.
 *
 *               Note that order of character area is not changed, only the
 *               order of the pointers.
 */

