
#include "call_em.h"
#include <grp.h>
#include <unistd.h>

extern char pri_device[];
extern int set_speed( int, enum how );
extern int read_defs(char *, char *);
extern int set_locks();
extern int zap_locks();
extern int set_devices(char *);
extern int check_locks(void);

int debug;
struct termios old, raw;

void usage(void);
char raw_char(void);
int display_char(unsigned char, int);
int query_modem(char *, char *);

int
main(int argc, char *argv[])
{
	time_t now;
	char *p;
	char getty_entry[30];
	int i;

	getty_entry[0] = '\0';
	i = 0;
	
	p = argv[0];	/* for gcc */
	while (i < argc) {
		if (!(p = argv[i++]))
			break;
		if (*p == '-') {
			switch (*++p) {
			case 's':
				if (!*++p)
					p = argv[i++];
				strncpy(getty_entry, p, sizeof(getty_entry));
			case 'd':
			case 'x':				  /* set the debug level */
				if (!*++p)
					p = argv[i++];
				debug = atoi(p);
				break;
			}
		} else if (*p == '?')
			usage();
	}
	set_devices(p);
	if (getty_entry[0] == '\0')
		strcpy(getty_entry, "T19200");	/* Set default */
	while (check_locks() != 0) {
		if (debug) {
			now = time(0L);
			fprintf(stderr, "%.24s: %s locks are set. Sleeping...\n",
					  ctime(&now), pri_device);
		}
		sleep(30);
	}
	if (set_locks() == 0) {
		ioctl(0, TCGETS, &old);
		raw = old;
		raw.c_cc[VMIN] = 0;
		raw.c_cc[VTIME] = 0;
		raw.c_lflag = old.c_lflag & ~(ICANON | ECHO);
		query_modem(pri_device, getty_entry);
	}
	zap_locks();
	time(&now);
	fprintf(stderr, "%.24s: %s Exiting\n", ctime(&now), pri_device);
	exit(0);
}

void
usage()
{
	fprintf(stdin, "uugetty: [-r] -t device entry\n");
	fprintf(stdin, "uugetty: -c gettydefs-type-file\n");
	fprintf(stdin, "uugetty: -q device\n");
	exit(0);
}

char
raw_char()
{
	char ch;

	ioctl(0, TCSETS, &raw);
	ch = getchar();
	ioctl(0, TCSETS, &old);
	return(( ch >= ' ' && ch <= '~' ) ? ch : '\0' );
}

/*
 * Put a char on stderr formatted as ^ or M- as required:
 * which == 0 put a char
 * which == 1 accumlate a line and insert a \n
 */
int
display_char(unsigned char ch, int which)
{
	int count = 0;
	static int total;
	char buf[80], *p;

	p = buf + total;
	if (ch > 0x7f) {
		*p++ = 'M';
		*p++ = '-';
		ch &= 0x7f;
	}
	if (ch < ' ') {
		if (ch == '\n')
			*p++ = ch;
		else {
			*p++ = '^';
			*p++ = ch + 0x40;
		}
	} else if (ch < 0x7F)
		*p++ = ch;
	else {
		*p++ = '^';
		*p++ = '?';
	}
	count = (p - buf) - total;
	if (which) {
		if ((ch == '\n') || ((total += count) >= 77)) {
			total = 0;
			*p++ = '\n';
			*p = '\0';
			fputs(buf, stderr);
		}
	} else {
		/* while( count++ < 4 ) *p++ = ' '; */
		*p = '\0';
		fputs(buf, stderr);
	}
	return (count);
}

int
query_modem(char *device, char *getty_entry)
{
	int fd, count, vee, ret;
	int flags;
	unsigned char ch1, ch2, ch3;
	unsigned char ch, *q, *r;
	char *p, buf[80];
	char *msg1 = "OK\n";
	char *msg2 = "...\n";
	struct termios term = {
		0,
		0,
		B19200 | CS8 | CREAD | HUPCL,
		0,
		N_TTY,
		{ 0x03, 0x1c, 0x08, 0x15, 0x04, 0x00, 0x00, 0x00, 0x11,
				  0x13, 0x1a, 0x00, 0x12, 0x0f, 0x17, 0x00, 0x00 },
	};
/*
 * BRKINT, IGNCR
 * OPOST, ONLCR
 * B19200, CS8, CREAD, CLOCAL, HUPCL
 * ISIG, ECHOK, ECHOCTL, ECHOKE
 *		0x0082, 0x0005, 0x0cbe, 0x0a21, 0,
 * VINTR VQUIT VERASE VKILL VEOF     VTIME    VMIN    VSWTC  VSTART
 * ETX,  FS,   BS,    NAK,  EOT,     NUL,     NUL,    NUL,   DC1,
 * ^c,   ^/,   ^h,    ^u,   ^d,      ^@,      ^@,     ^@,    ^q,
 *
 *       VSTOP VSUSP VEOL   VREPRINT VDISCARD VWERASE VLNEXT VEOL2
 *       DC3,  SUB,  NUL,   DC2,     SI,      ETB,    NUL    NUL
 *       ^s,   ^z,   ^@,    ^r,      ^q,      ^w,     ^@     ^q
 */

	ioctl(0, TCSETS, &raw);
	if ((fd = open(device, O_RDWR )) < 0) {
		fprintf(stderr, "Unable to open %s, %s\n",
				  device, sys_errlist[errno]);
		return (-1);
	}
	ioctl( fd, TCSETSW, &term );
	fprintf(stderr, "Port '%s' open. fd = %d.\n", device, fd );
	sleep(1);

	/* Toggle DTR just for laughs */
	flags = TIOCM_DTR | TIOCM_RTS;
	ioctl(fd, TIOCMBIC, &flags); /* Turn off flags */
	fprintf( stderr, "Use '+' to send escape sequence.\n" );
	sleep(1);
	ioctl(fd, TIOCMBIS, &flags); /* Turn on flags */
	sleep(1);

	p = buf;
	ch3 = ch1 = '\0';
	vee = 0;
	count = -1;
	fprintf( stderr, "Use '/' to send 'AT&V0'.\n" );
	fprintf( stderr, "Use '*' to send Dumb Modem sequence.\n" );
	fprintf( stderr, "\nESC to quit, Enter command > ");
	while (( ch = getchar()) != '\033') {
		if( ch != ch1 ) {
			if( ch >= ' ' && ch <= '~' ) {
				if( ch >= 'a' && ch <= 'z' ) ch -= ' ';
				*p++ = ch;
				putc( ch, stderr );
			}
			else if (( ch == 0x7f ) && (p > buf)) {
				p--;
				fputs( "\b \b", stderr);
			}
			else if ( ch == '\n') {
				*p++ = '\r';
				*p++ = '\n';
				*p = '\0';
				write(fd, &buf, p - buf);
				p = buf;
				putc( '\n', stderr );
				count = 20000;
			}
			if ( ch == '+') {
				write(fd, "+++", 3);
				p = buf;
				fprintf( stderr, "\nEnter command > ");
			}
			else if( ch == '*' ) {
				int i, j;
				for( i = 40, j = 0; i; i-- ) {
					write( fd, "A", 1 );
					for( count = 2000; count; count-- );
					if(( read( fd, &ch, 1 )) == 1 ) {
						display_char( ch, 0 );
						if( ch == 'A') j++;
						if( j == 3 ) break;
						else j= 0;
					}
				}
				if( i ) write( fd, "AT***\r\n", 7 );
				else fputs( "Failed.\n", stderr );
			}
			else if ( ch == '/') {
				write(fd, "AT&V\r\n", 7);
				fputs( " AT&V\n", stderr );
				vee = 1;
				count = 20000;
			}
			ch1 = ch;
		}
		q = msg1;
		r = msg2;
		while(( ret = read( fd, &ch2, 1 )) > 0 ) {
			if( ch2 == '\r' || ( ch2 == '\n' && ch3 == '\n' )) continue;
			display_char(ch2, 0);
			ch3 = ch2;
			if( ch2 == *q ) q++;
			else q = msg1;
			if( ch2 == *r ) r++;
			else r = msg2;
			ch = '\0';
			if( ! *q ) {
				vee = 0;
				break;
			}
			if( ! *r ) break;
		}
		if( count > 0 ) count--;
		if( count == 0 || ( ch3 == '\n' && ret )) {
			if( vee )
					fprintf( stderr, "<cr> to continue." );
			else
					fprintf( stderr, "ESC to quit, Enter command > ");
			count = -1;
		}
	}
	fprintf(stderr, "\nStopped ..\n");
	ioctl(fd, TCFLSH, 2);
	close(fd);
	ioctl(0, TCSETS, &old);
	return (0);
}
